##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
This module implements pytest plugin for Sphinx doc tests.

In a nutshell, it uses the pytest `pytest_collect_file()` plugin hook
to recognize the Sphinx Makefile. Then it does a quick and dirty parse
of that Makefile to extract the command Sphinx is using to run the
doctests, which it recognizes by being the first command in the
Makefile target named by `SPHINX_DOCTEST_TARGET`. The parser is
able to handle simple Makefile variable expansion, though not currently
nested variables so don't do that.

The mechanics of the pytest plugin mechanism are such that the Makefile
is wrapped with a subclass of :class:`pytest.File`, :class:`SphinxMakefile`,
which implements the `collect` method to yield a subclass of :class:`pytest.Item`
called :class:`SphinxItem`, that in turn implements a few methods to run the
test and report the result. The bulk of the code in running the test is parsing
the output to look for errors, and thus decide whether all the doctests passed,
or not.

The drawback of this whole setup is of course some extra complexity.
The advantage is that (a) whatever the Makefile does is what this plugin
should do, for running the command, as long as the command is the first
(and only significant) thing that occurs in the target, and (b) if there ends up
being more than one Makefile, it should all continue to work.
"""
import os
import re
import pathlib
import subprocess

#
import pytest

__author__ = 'Dan Gunter'

# Modify this to match the target in the
# Sphinx Makefile for running the doctests
SPHINX_DOCTEST_TARGET = "doctest"

# Modify this for non-standard sphinx-build commands.
# Knowing the Sphinx build command is needed to replicate it
SPHINX_BUILD = "sphinx-build"

g_sphinx_warn_file = None


def pytest_collect_file(parent, path):
    global g_sphinx_warn_file
    if path.ext == "" and path.basename == "Makefile":
        if file_contains("sphinx", str(path)):
            makefile = SphinxMakefile(path, parent)
            g_sphinx_warn_file = makefile.warnings_file
            return makefile
    elif g_sphinx_warn_file is not None and path.basename == g_sphinx_warn_file:
        return SphinxWarnings(path, parent)


def file_contains(s, path):
    result = False
    with open(path) as f:
        for line in f:
            if s in line:
                result = True
                break
    return result


class SphinxWarnings(pytest.File):
    def collect(self):
        path = pathlib.Path(self.fspath)  # convert to std
        yield SphinxWarningsItem('Sphinx warnings', self, path)


class SphinxWarningsItem(pytest.Item):
    def __init__(self, name, parent, path: pathlib.Path):
        super().__init__(name, parent)
        self._path = path
        self._warnings = None

    def runtest(self):
        if self._path.stat().st_size > 0:
            with self._path.open() as f:
                self._warnings = f.read()
            raise SphinxHadErrors

    def repr_failure(self, excinfo):
        """This is called when self.runtest() raises an exception.
        """
        return f"Build had warnings and/or errors:\n{self._warnings}"

    def reportinfo(self):
        summary = f"Sphinx doc warnings in {self._path}"
        return f"doc build warnings in {self._path}", 0, summary


class SphinxMakefile(pytest.File):

    # simple way to find variables in a Makefile
    makefile_var_defn = re.compile(r"\s*([A-Za-z_]+)\s*=(.*)")

    def collect(self):
        cmd = self._get_doctest_command()
        wd = os.path.dirname(self.fspath)
        if cmd is not None:
            yield SphinxDoctestItem('Sphinx doctests', self, wd, cmd)

    @property
    def warnings_file(self):
        """Get warnings and errors output file, if any, from the Sphinx Makefile.
        """
        mkvars = {}  # variables defined in Makefile
        with self.fspath.open() as makefile:
            for line in makefile:
                s = line.strip()
                # look for definition of Sphinx options
                if s.startswith("SPHINXOPTS"):
                    opts = s[s.find('=') + 1 :].strip()
                    # return value of '-w' option, if found
                    m = re.match("-w\s*['\"]?([^\"' \t]+)", opts)
                    return m.group(1) if m else None

    def _get_doctest_command(self):
        result = None
        mkvars = {}  # variables defined in Makefile
        with self.fspath.open() as makefile:
            cmd = ""
            state = "pre"
            for line in makefile:
                s = line.strip()
                # look for 'doctest' target
                if state == "pre":
                    if s.startswith(SPHINX_DOCTEST_TARGET):
                        state = "next"
                    else:
                        # look for a NAME = value definition, save it in 'mkvars'
                        m = self.makefile_var_defn.match(s)
                        if m is not None:
                            value = m.group(2).strip()
                            expanded_value = self._expand(value, mkvars)
                            mkvars[m.group(1)] = expanded_value
                # primed for this to be the doctest command
                elif state == "next":
                    # use 'mkvars' to expand out command
                    if s.endswith("\\"):
                        cmd += s[:-1] + " "
                    else:
                        cmd += s
                        expanded = self._expand(cmd, mkvars)
                        if SPHINX_BUILD not in expanded:
                            cmd = ""
                        else:
                            result = expanded
                            state = "done"
                elif state == "done":
                    break
            if result is None and state == "next":
                raise ValueError(f"EOF while parsing doctest command '{cmd}'")
        return result

    @staticmethod
    def _expand(s, mkvars):
        """Expand out the variable refs by making them look like format vars, then
        simply using string formatting.
        """
        fstr, n = re.subn(r"\$\(([A-Za-z_]+)\)", "{\\1}", s)
        return fstr.format(**mkvars)


class SphinxDoctestItem(pytest.Item):

    __executed = False

    def __init__(self, name, parent, wd, cmd):
        """New Sphinx doctest item.

        Args:
            name (str): Item name
            parent (pytest.File): Parent item
            wd (str): working directory
            cmd (str): Command to run
        """
        super().__init__(name, parent)
        self._sess = parent.session
        # working directory and command
        self.wd, self.cmd = wd, cmd
        self.successes, self.failures = None, None
        self.total_successes, self.total_failures = -1, -1
        self.failure_list = []

    def runtest(self):
        """Run the Sphinx doctest.
        """
        self._insert_items_at = self._sess.items.index(self) + 1
        #  print(f"\n@@ SphinxDoctestItem.run(): session items: {self._sess.items}.\n"
        #      f"Insert at: {self._insert_items_at}\n")
        if self.__executed:
            return
        self.__executed = True
        old_d = os.getcwd()
        print(f"doctest command: {self.cmd}")
        try:
            os.chdir(self.wd)
            args = self.cmd.split()
            try:
                proc = subprocess.run(
                    args,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                    timeout=300,
                    encoding='utf-8',
                )
                self._parse_output(proc.stdout)
            except Exception as exc:
                print(f"Unexpected failure in Sphinx doctest command: {exc}")
                raise SphinxCommandFailed(self.cmd, str(exc))
        finally:
            os.chdir(old_d)

    def repr_failure(self, excinfo):
        """This is called when self.runtest() raises an exception.
        """
        if not isinstance(excinfo.value, SphinxHadErrors):
            return None
        header = f"doctest execution failed: {self.total_failures} failures"
        body = "\n".join(self.failure_list)
        return header + "\n" + body

    def reportinfo(self):
        summary = f"Sphinx doctests in {self.wd}"
        return f"doctests in {self.wd}", 0, summary

    def _parse_output(self, output):
        """Parse output of Sphinx doctest.
        """

        def _msgblock(messages):
            lines, indent, marker = [], None, " ** "
            for m in messages:
                first_line = marker
                indent = " " * len(first_line)
                for s in m.split("\n"):
                    if first_line:
                        first_line += s
                        lines.append(first_line)
                        first_line = None
                    else:
                        lines.append(indent + s)
            return "\n".join(lines)

        state, failed_msg = "ok", ""
        cur_doc, n_passed, n_failed = "", -1, -1
        for line in output.split('\n'):
            line = line.rstrip()
            # print(f"@@ doctest: {line}")
            try:
                if state == "ok":
                    if line.startswith('****'):
                        # start of a failure message
                        state = "fail"
                        failed_msg = ""
                    elif line.startswith("Document:"):
                        # start of a new doc
                        _, cur_doc = line.split(None, 1)
                        n_passed, n_failed = 0, 0
                    elif re.match(r"\d+ passed and \d+ failed", line):
                        # summary of pass/fail for a doc
                        m = re.match(r"(\d+) passed and (\d+) failed", line)
                        n_passed, n_failed = int(m.group(1)), int(m.group(2))
                    elif ("Test passed" in line) or ("Test Failed" in line):
                        # final line of report for a doc
                        if n_passed > 0:
                            self.tests_ok(cur_doc, n_passed)
                        cur_doc = None
                elif state == "fail":
                    if line.startswith("****"):
                        # end of failure record
                        self.test_failed(cur_doc, failed_msg)
                        state = "ok"
                        # do not reset 'cur_doc', there may be more..
                    else:
                        # middle of failure record
                        if failed_msg:
                            failed_msg += "\n"
                        failed_msg += line
            except Exception as exc:
                print(f"Parse error on line '{line}' :: {exc}")
                raise
        # print(f"@@ done doctest output")

    def tests_ok(self, context, num):
        # print(f"Tests in {context}: {num} OK")
        test_name = "Sphinx doctest in: {context}"
        if num > 1:
            test_name += " ({i})"
        for i in range(num, 0, -1):
            success_item = SphinxDoctestSuccess(
                test_name.format(**locals()), self.parent
            )
            self._sess.items.insert(self._insert_items_at, success_item)
        self._sess.testscollected += num

    def test_failed(self, context, failure):
        test_name = "Sphinx doctest in: {context}.rst"
        item = SphinxDoctestFailure(test_name.format(**locals()), self.parent, failure)
        self._sess.items.insert(self._insert_items_at, item)
        self._sess.testscollected += 1


class SphinxDoctestSuccess(pytest.Item):
    def runtest(self):
        return


class SphinxDoctestFailure(pytest.Item):
    def __init__(self, name, parent, details):
        super().__init__(name, parent)
        self.details = details

    def runtest(self):
        raise SphinxHadErrors()

    def repr_failure(self, excinfo):
        return "doctest failure:\n" + self.details

    def reportinfo(self):
        return f"doctest in {self.name}", 0, f"FAILED {self.name}"


class SphinxCommandFailed(Exception):
    def __init_(self, cmd, msg):
        super().__init__(f"Sphinx doctest command ({cmd}) failed: {msg}")


class SphinxHadErrors(Exception):
    pass



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Writer of the data into the GAMS file
"""

import numpy as np
import matplotlib.pyplot as plt
from . import BasisFunctions, DataImport

Combination = False
R = 8.314472


def writeExp(textFile, Combination=False):
    """
    Writes into the GDX file the basis function parameters
    """
    coeffs = BasisFunctions.coeffs

    i = 1
    for d, t, l, m in coeffs:
        textFile.write("d('%d') = %i ;\n " % (i, d))
        textFile.write("t('%d') = %.6f ;\n " % (i, t))
        textFile.write("l('%d') = %i ;\n " % (i, l))
        i = i + 1


def PVTdt(textFile, DataToWrite, Combination=False, PlotData=False):
    """ Imports P-V-T data into the GAMS document, to be written in :func:'GamsWrite'

    :param textFile: Gams File written to.
    :type textFile: str.
    :param DataToWrite: Data prepared for the GAMS file.
    :type DataToWrite: array.
    :returns: void.
    """

    IsothermIndexes = DataImport.isothermIndex

    i = 1
    for x in DataToWrite:
        if Combination:
            textFile.write("z('%s', '%d') = %f ;\n" % ("PVT", i, x[2]))
        else:
            textFile.write("z('%d') = %f ;\n" % (i, x[2]))
        i += 1
    i = 1

    for x in DataToWrite:
        if Combination:
            textFile.write("delta('%s','%d') = %f;\n" % ("PVT", i, x[0]))
            textFile.write("tau('%s','%d') = %f;\n" % ("PVT", i, x[1]))
        else:
            textFile.write("delta('%d') = %f;\n" % (i, x[0]))
            textFile.write("tau('%d') = %f;\n" % (i, x[1]))

        i += 1
    i = 1

    InSat(textFile, DataImport.InSatValues,Combination)

    if PlotData:
        x, y, z = [], [], []
        for xs in DataToWrite:
            x.append(xs[0])
            y.append(xs[1])
            z.append(xs[2])

        fig = plt.figure()
        ax = fig.add_subplot(111, projection="3d")
        ax.scatter(x, y, z, edgecolor="k", c=y, marker="o", s=25)
        ax.xaxis.set_ticks(np.arange(min(x), max(x), (max(x) - min(x)) / 4))
        ax.yaxis.set_ticks(np.arange(min(y), max(y), (max(y) - min(y)) / 5))
        ax.zaxis.set_ticks(np.arange(min(z), max(z), (max(z) - min(z)) / 5))
        ax.set_xlabel("delta", labelpad=20, fontsize=20)
        ax.set_ylabel("tau", labelpad=20, fontsize=20)
        ax.set_zlabel("z", labelpad=10, fontsize=20)
        plt.show()


def CVdt(textFile, DataToWrite, Combination=False, PlotData=False):
    """ Imports Isochoric Heat Capacity (CV) data into the GAMS document

    :param textFile: Gams File written to.
    :type textFile: str.
    :param DataToWrite: Data prepared for the GAMS file.
    :type DataToWrite: array.
    :returns: void.
    """

    i = 1

    for x in DataToWrite:
        BasisFunctions.iTT(x[0], x[1])
        itt_val = -BasisFunctions.itt_val
        val = x[2]
        if Combination:
            textFile.write("z('%s', '%d') = %f ;\n" % ("CV", i, val))
            textFile.write("itt('%s', '%d') = %f ;\n" % ("CV", i, itt_val))
            textFile.write("delta('%s', '%d') = %.13f ;\n" % ("CV", i, x[0]))
            textFile.write("tau('%s','%d') = %.13f ;\n" % ("CV", i, x[1]))
        else:
            textFile.write("z('%d') = %f ;\n" % (i, val))
        i += 1

    if PlotData:
        x, y, z = [], [], []
        for xs in DataToWrite:
            x.append(xs[0])
            y.append(xs[1])
            z.append(xs[2])

        fig = plt.figure()
        ax = fig.add_subplot(111, projection="3d")
        ax.scatter(x, y, z, edgecolor="k", c="none", marker="o")
        ax.set_xlabel("delta")
        ax.set_ylabel("tau")
        ax.set_zlabel("CV")
        plt.show()


def CPdt(textFile, DataToWrite, Combination=False, PlotData=False): 
    """ Imports Isobaric Heat Capacity(CP) data into the GAMS document

    :param textFile: Gams File written to.
    :type textFile: str.
    :param DataToWrite: Data prepared for the GAMS file.
    :type DataToWrite: array.
    :returns: void.

    """

    i = 1
    for x in DataToWrite:
        BasisFunctions.iTT(x[0], x[1])
        itt_val = -BasisFunctions.itt_val
        val = x[2] - itt_val
        if Combination:
            textFile.write("z('%s', '%d') = %.13f ;\n" % ("CP", i, val))
        else:
            textFile.write("z('%d') = %.13f ;\n" % (i, val))
        i += 1
    i = 1
    for x in DataToWrite:

        if Combination:
            textFile.write("delta('%s', '%d') = %.13f ;\n" % ("CP", i, x[0]))
            textFile.write("tau('%s','%d') = %.13f ;\n" % ("CP", i, x[1]))
        else:
            textFile.write("delta( '%d') = %.13f ;\n" % (i, x[0]))
            textFile.write("tau('%d') = %.13f ;\n" % (i, x[1]))

        i += 1

    if PlotData:
        x, y, z = [], [], []
        for xs in DataToWrite:
            x.append(xs[0])
            y.append(xs[1])
            z.append(xs[2])

        fig = plt.figure()
        ax = fig.add_subplot(111, projection="3d")
        ax.scatter(x, y, z, edgecolor="k", c=z, marker="o", s=30)
        ax.xaxis.set_ticks(np.arange(min(x), max(x), (max(x) - min(x)) / 4))
        ax.yaxis.set_ticks(np.arange(min(y), max(y), (max(y) - min(y)) / 5))
        ax.zaxis.set_ticks(np.arange(min(z), max(z), (max(z) - min(z)) / 5))
        ax.set_xlabel("delta", labelpad=20, fontsize=20)
        ax.set_ylabel("tau", labelpad=20, fontsize=20)
        ax.set_zlabel("Isobaric heat capacity", labelpad=10, fontsize=20)

        # ax.set_title('Molecule isobaric heat capacity data')
        plt.show()


def SNDdt(textFile, DataToWrite, Combination=False, PlotData=False):
    """ Imports Speed of Sound (SND) data into the GAMS document

    :param textFile: Gams File written to.
    :type textFile: str.
    :param DataToWrite: Data prepared for the GAMS file.
    :type DataToWrite: array.
    :returns: void.

    """
    i = 1
    for x in DataToWrite:
        if Combination:
            textFile.write("z('%s', '%d') = %.13f ;\n" % ("SND", i, x[2]))
        else:
            textFile.write("z('%d') = %.13f ;\n" % (i, x[2]))
        i += 1
    i = 1
    for x in DataToWrite:
        BasisFunctions.iTT(x[0], x[1])
        itt_val = BasisFunctions.itt_val
        BasisFunctions.drd(x[0], x[1])
        if Combination:
            textFile.write("itt('%s','%d') = %.13f ;\n" % ("SND", i, itt_val))
            textFile.write("delta('%s', '%d') = %.13f ;\n" % ("SND", i, x[0]))
            textFile.write("tau('%s','%d') = %.13f ;\n" % ("SND", i, x[1]))
        else:
            textFile.write("itt('%s','%d') = %.13f ;\n" % (i, itt_val))
            textFile.write("delta('%s', '%d') = %.13f ;\n" % (i, x[0]))
            textFile.write("tau('%s','%d') = %.13f ;\n" % (i, x[1]))
        i += 1


    if PlotData:
        x, y, z = [], [], []
        for xs in DataToWrite:
            x.append(xs[0])
            y.append(xs[1])
            z.append(xs[2])

        fig = plt.figure()
        ax = fig.add_subplot(111, projection="3d")
        ax.scatter(x, y, z, edgecolor="k", c=z, marker="o", s=30)

        fig = plt.figure()
        ax = fig.add_subplot(111, projection="3d")
        ax.scatter(x, y, z, edgecolor="k", c=z, marker="o")
        ax.xaxis.set_ticks(np.arange(min(x), max(x), (max(x) - min(x)) / 4))
        ax.yaxis.set_ticks(np.arange(min(y), max(y), (max(y) - min(y)) / 5))
        ax.zaxis.set_ticks(np.arange(min(z), max(z), (max(z) - min(z)) / 5))
        ax.set_xlabel("delta", labelpad=20, fontsize=20)
        ax.set_ylabel("tau", labelpad=20, fontsize=20)
        ax.set_zlabel("Speed of sound", labelpad=10, fontsize=20)

        plt.show()


def Crit(textFile, DataToWrite, Combination=False):
    """ Import Critical data points"""
    BasisFunctions.drd(1, 1)
    drd_vals = BasisFunctions.drd_vals

    i = 1
    j = 1
    for y in drd_vals:
        if Combination:
            textFile.write(
                "crit('%s', '%d', '%d', '%s') = %f ;\n" % ("PVT", i, j, "drd", y)
            )
        else:
            textFile.write("crit('%d', '%d', '%s') = %f ;\n" % (i, j, "drd", y))
        j += 1
    i = 1
    j = 1
    BasisFunctions.d2rd(1, 1)
    d2rd_vals = BasisFunctions.d2rd_vals
    for y in d2rd_vals:
        if Combination:
            textFile.write(
                "crit('%s', '%d', '%d', '%s') = %f ;\n" % ("PVT", i, j, "d2rd", y)
            )
        else:
            textFile.write("crit('%d', '%d', '%s') = %f ;\n" % (i, j, "d2rd", y))
        j += 1
    i = 1
    j = 1
    BasisFunctions.d3rd(1, 1)
    d3rd_vals = BasisFunctions.d3rd_vals
    for y in d3rd_vals:
        if Combination:
            textFile.write(
                "crit('%s', '%d', '%d', '%s') = %f ;\n" % ("PVT", i, j, "d3rd", y)
            )
        else:
            textFile.write("crit('%d', '%d', '%s') = %f ;\n" % (i, j, "d3rd", y))
        j += 1
    i = 1


def InSat(textFile, DataToWrite, Combination=False):
    """Import saturation density values"""
    i = 1
    for x in DataToWrite:
        j = 1
        BasisFunctions.d3rd(x[0], x[1])
        BasisFunctions.d4rd(x[0], x[1])
        BasisFunctions.d5rd(x[0], x[1])
        d3rd_vals = BasisFunctions.d3rd_vals
        d4rd_vals = BasisFunctions.d4rd_vals
        d5rd_vals = BasisFunctions.d5rd_vals
        for x, y, z in zip(d3rd_vals, d4rd_vals, d5rd_vals):
            val = 12 * x + 8 * y + z
            if Combination:
                textFile.write(
                    "InSat('%s', '%d', '%d', '%s') = %f ;\n"
                    % ("PVT", i, j, "d3rd", val)
                )
            else:
                textFile.write("InSat('%d', '%d', '%s') = %f ;\n" % (i, j, "d3rd", val))
            j += 1
        i += 1



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Standard IDAES Gibbs reactor model.
"""
# Import Pyomo libraries
from pyomo.environ import Reals, Var
from pyomo.common.config import ConfigBlock, ConfigValue, In

# Import IDAES cores
from idaes.core import (ControlVolume0DBlock,
                        declare_process_block_class,
                        EnergyBalanceType,
                        MomentumBalanceType,
                        UnitModelBlockData,
                        useDefault)
from idaes.core.util.config import is_physical_parameter_block
from idaes.core.util.misc import add_object_reference

__author__ = "Jinliang Ma, Andrew Lee"


@declare_process_block_class("GibbsReactor")
class GibbsReactorData(UnitModelBlockData):
    """
    Standard Gibbs Reactor Unit Model Class

    This model assume all possible reactions reach equilibrium such that the
    system partial molar Gibbs free energy is minimized.
    Since some species mole flow rate might be very small,
    the natural log of the species molar flow rate is used.
    Instead of specifying the system Gibbs free energy as an objective
    function, the equations for zero partial derivatives of the grand function
    with Lagrangian multiple terms with repect to product species mole flow
    rates and the multiples are specified as constraints.
    """
    CONFIG = ConfigBlock()
    CONFIG.declare("dynamic", ConfigValue(
        domain=In([False]),
        default=False,
        description="Dynamic model flag - must be False",
        doc="""Gibbs reactors do not support dynamic models, thus this must be
False."""))
    CONFIG.declare("has_holdup", ConfigValue(
        default=False,
        domain=In([False]),
        description="Holdup construction flag",
        doc="""Gibbs reactors do not have defined volume, thus this must be
False."""))
    CONFIG.declare("energy_balance_type", ConfigValue(
        default=EnergyBalanceType.useDefault,
        domain=In(EnergyBalanceType),
        description="Energy balance construction flag",
        doc="""Indicates what type of energy balance should be constructed,
**default** - EnergyBalanceType.useDefault.
**Valid values:** {
**EnergyBalanceType.useDefault - refer to property package for default
balance type
**EnergyBalanceType.none** - exclude energy balances,
**EnergyBalanceType.enthalpyTotal** - single enthalpy balance for material,
**EnergyBalanceType.enthalpyPhase** - enthalpy balances for each phase,
**EnergyBalanceType.energyTotal** - single energy balance for material,
**EnergyBalanceType.energyPhase** - energy balances for each phase.}"""))
    CONFIG.declare("momentum_balance_type", ConfigValue(
        default=MomentumBalanceType.pressureTotal,
        domain=In(MomentumBalanceType),
        description="Momentum balance construction flag",
        doc="""Indicates what type of momentum balance should be constructed,
**default** - MomentumBalanceType.pressureTotal.
**Valid values:** {
**MomentumBalanceType.none** - exclude momentum balances,
**MomentumBalanceType.pressureTotal** - single pressure balance for material,
**MomentumBalanceType.pressurePhase** - pressure balances for each phase,
**MomentumBalanceType.momentumTotal** - single momentum balance for material,
**MomentumBalanceType.momentumPhase** - momentum balances for each phase.}"""))
    CONFIG.declare("has_heat_transfer", ConfigValue(
        default=False,
        domain=In([True, False]),
        description="Heat transfer term construction flag",
        doc="""Indicates whether terms for heat transfer should be constructed,
**default** - False.
**Valid values:** {
**True** - include heat transfer terms,
**False** - exclude heat transfer terms.}"""))
    CONFIG.declare("has_pressure_change", ConfigValue(
        default=False,
        domain=In([True, False]),
        description="Pressure change term construction flag",
        doc="""Indicates whether terms for pressure change should be
constructed,
**default** - False.
**Valid values:** {
**True** - include pressure change terms,
**False** - exclude pressure change terms.}"""))
    CONFIG.declare("property_package", ConfigValue(
        default=useDefault,
        domain=is_physical_parameter_block,
        description="Property package to use for control volume",
        doc="""Property parameter object used to define property calculations,
**default** - useDefault.
**Valid values:** {
**useDefault** - use default package from parent model or flowsheet,
**PropertyParameterObject** - a PropertyParameterBlock object.}"""))
    CONFIG.declare("property_package_args", ConfigBlock(
        implicit=True,
        description="Arguments to use for constructing property packages",
        doc="""A ConfigBlock with arguments to be passed to a property block(s)
and used when constructing these,
**default** - None.
**Valid values:** {
see property package for documentation.}"""))

    def build(self):
        """
        Begin building model (pre-DAE transformation).

        Args:
            None

        Returns:
            None
        """
        # Call UnitModel.build to setup dynamics
        super(GibbsReactorData, self).build()

        # Build Control Volume
        self.control_volume = ControlVolume0DBlock(default={
                "dynamic": self.config.dynamic,
                "property_package": self.config.property_package,
                "property_package_args": self.config.property_package_args})

        self.control_volume.add_state_blocks(has_phase_equilibrium=False)

        self.control_volume.add_total_element_balances()

        self.control_volume.add_energy_balances(
            balance_type=self.config.energy_balance_type,
            has_heat_transfer=self.config.has_heat_transfer)

        self.control_volume.add_momentum_balances(
            balance_type=self.config.momentum_balance_type,
            has_pressure_change=self.config.has_pressure_change)

        # Add Ports
        self.add_inlet_port()
        self.add_outlet_port()

        # Add performance equations
        # Add Lagrangian multiplier variables
        self.lagrange_mult = Var(self.flowsheet().config.time,
                                 self.config.property_package.element_list,
                                 domain=Reals,
                                 initialize=100,
                                 doc="Lagrangian multipliers")

        # Use Lagrangian multiple method to derive equations for Out_Fi
        # Use RT*lagrange as the Lagrangian multiple such that lagrange is in
        # a similar order of magnitude as log(Yi)

        @self.Constraint(self.flowsheet().config.time,
                         self.config.property_package.phase_list,
                         self.config.property_package.component_list,
                         doc="Gibbs energy minimisation constraint")
        def gibbs_minimization(b, t, p, j):
            # Use natural log of species mole flow to avoid Pyomo solver
            # warnings of reaching infeasible point
            return 0 == (
                b.control_volume.properties_out[t].gibbs_mol_phase_comp[p, j] +
                sum(b.lagrange_mult[t, e] *
                    b.control_volume.properties_out[t].
                    config.parameters.element_comp[j][e]
                    for e in b.config.property_package.element_list))

        # Set references to balance terms at unit level
        if (self.config.has_heat_transfer is True and
                self.config.energy_balance_type != EnergyBalanceType.none):
            add_object_reference(self,
                                 "heat_duty",
                                 self.control_volume.heat)
        if (self.config.has_pressure_change is True and
                self.config.momentum_balance_type != MomentumBalanceType.none):
            add_object_reference(self,
                                 "deltaP",
                                 self.control_volume.deltaP)

    def _get_performance_contents(self, time_point=0):
        var_dict = {}
        if hasattr(self, "heat_duty"):
            var_dict["Heat Duty"] = self.heat_duty[time_point]
        if hasattr(self, "deltaP"):
            var_dict["Pressure Change"] = self.deltaP[time_point]

        return {"vars": var_dict}



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Parses and prints the solutions of the multiparameter equation of state solution
"""
# stdlib
import re
# pkg
from . import BasisFunctions

indexes = []
betas = []


def parser(filename, num=2):
    """
    Parse solution files for the muliparameter equation of state
    """
    global indexes, betas
    print("Parsing %s" % filename)
    dataFile = open(filename)

    # Y variables regex
    regexY = r"(?<=(VARIABLE y.L))(\s*)(([0-9,. \s]+)(\n))"
    regexB = r"(?<=(VARIABLE beta.L))(\s*)(([-E0-9,. \s]+)(\n))"

    dataFile = open(filename)

    Yb = re.findall(regexY, dataFile.read())

    indexes = []
    for ybi in Yb:
        indexes_sub = []
        ybi = " ".join(ybi[2].split())
        groups = ybi.split(",")
        for i in groups:
            trimmed = i.strip()
            splits = trimmed.split(" ")
            indexes_sub.append(int(splits[0]))
            if len(splits) > 3:
                indexes_sub.append(int(splits[2]))
        indexes.append(indexes_sub)

    dataFile2 = open(filename)
    Yb = re.findall(regexB, dataFile2.read())

    betas = []
    ind = 0
    for ybi in Yb:
        betas_sub = []
        ybi = " ".join(ybi[2].split())
        groups = ybi.split(",")
        for i in groups:
            i = i.strip()
            splits = i.split(" ")
            if len(splits) <= 2:
                beta = splits[1]
                index = int(splits[0])
            else:
                beta = splits[1]
                index = int(splits[0])
                if index in indexes[ind]:
                    betas_sub.append(beta)
                beta = splits[3]
                index = int(splits[2])

            if index in indexes[ind]:
                betas_sub.append(beta)
        ind = ind + 1
        betas.append(betas_sub)


def getIndexes():
    """
    Returns indexes of the basis function terms
    """
    global indexes
    return indexes


def getBetas():
    """
    Returns the weights of the basis functions
    """
    global betas
    return betas


def writeTerm(index):
    """
    Writes the basis function term with the given index
    """
    global coeffs
    coeffs = []
    eqtn = ""
    term_index = 0
    for i in range(1, 9):
        for j in range(1, 13):  # 12
            term_index += 1
            coeffs.append([i, j / 8.0, 0])
            if index == term_index:
                eqtn = eqtn + " %s^%d * %s^%.3f" % ("D", i, "T", j / 8.0)
    for i in range(1, 6):
        for j in range(1, 24):  # 24
            term_index += 1
            coeffs.append([i, j / 8.0, 1])
            if index == term_index:
                eqtn = eqtn + " %s^%d * %s^%.3f * exp(-D^1)" % ("D", i, "T", j / 8.0)
    for i in range(1, 6):
        for j in range(1, 30):  # 24
            term_index += 1
            coeffs.append([i, j / 8.0, 2])
            if index == term_index:
                eqtn = eqtn + " %s^%d * %s^%.3f * exp(-D^2)" % ("D", i, "T", j / 8.0)
    for i in range(2, 5):
        for j in range(24, 38):  # 38
            term_index += 1
            coeffs.append([i, j / 2.0, 3])
            if index == term_index:
                eqtn = eqtn + " %s^%d * %s^%.3f * exp(-D^3)" % ("D", i, "T", j / 8.0)
    return eqtn


def writeEquation(Y, Beta =None):
    """
    Write full multiparameter equation
    """
    global coeffs
    global indexes

    eqtn = "Equation for Helmholtz Energy \n ="
    coeffs = BasisFunctions.coeffs
    indexes = Y
    lastInd = indexes[-1]
    for ind, b in zip(indexes, Beta):
        d, t, c, m = coeffs[ind-1]
        if not c == 0:
            eqtn = eqtn + "%f * %s^%d * %s^%.3f * exp(-D^%d)" % (
                        b,
                        "D",
                        d,
                        "T",
                        t,
                        c
                    )
        else:
            eqtn = eqtn + "%f * %s^%d * %s^%.3f" % (
                        b,
                        "D",
                        d,
                        "T",
                        t
                    )
        if not ind == lastInd:
            eqtn = eqtn + " +"

    print(eqtn)


##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Run ALAMO.
"""
import sys
import collections
import numpy as np
import os

#
from idaes.surrogate import alamopy
from idaes.surrogate.alamopy import almerror
from idaes.surrogate.alamopy.writethis import writethis
from idaes.surrogate.alamopy.multos import deletefile


def doalamo(xdata, zdata, **kwargs):
    """
    Warning: doalamo is deprecated. 
    please use alamopy.doalamo.alamo(xdata, zdata, **kwargs)
    """
    return alamo(xdata, zdata, **kwargs)


def alamo(xdata, zdata, **kwargs):
    """ [almmodel] = doalamo(xdata,zdata, xvaldata, zvaldata,addopt=vals)

    Args:
        xdata (numpy.array or list[real])
        zdata (numpy.array or list[real)
        kwargs: Additional options may be specified and will be applied
                to the .alm
          -  example -  monomialpower=(1,2,3,4) 
          -  xlabels       : labels given to input variables
          -  zlabels       : labels given to outputs
          -  xval          : validaiton data for alamo
          -  zval          : response validation data for alamo
          -  modeler       : modeler value used in alamo
          -  solvemip      : force alamo to solve mip if gams is availible
          -  linfcns       : 0-1 option to include linear transformations
          -  expfcns       : 0-1 option to include exponential transformations
          -  logfcns       : 0-1 option to include logarithmic transformations
          -  sinfcns       : 0-1 option to include sine transformations
          -  cosfcns       : 0-1 option to include cosine transformations
          -  monomialpower : list of monomial powers
          -  multi2power   : list of binomial powers
          -  multi3power   : list of trinomials
          -  ratiopower    : list of ratio powers
          -  screener      : screening method
          -  almname       : specify a name for the .alm file
          -  savescratch   : saves .alm and .lst
          -  savetrace     : saves trace file
          -  expandoutput  : add a key to the output dictionary for the output
                             (must be on for inputs(outputs?#Engle)>1)
          -  almopt        : direct text appending
                             the option almopt=<file> will append a file to the
                             end of the .alm and can be used to facilitate
                             direct access to the .alm (no current checks)
          -  loo           : leave one out evaluation
          -  lmo           : leave many out evaluation
    Returns:
        dict: An ALAMO model with the following keys
          -  'model'    : algebraic form of model
          -  'f(model)' : a callable lambda function
          -   Syntac is depended on expandout
               syntax => almmodel['f(model)']['out'](inputs,sep,by,comma)
                         almmodel['f(model)'](inputs,sep,by,comma)
          -  'ssr'      : SSE on training set provided
          -  'R2'       : R2 on training set provided
          -  'ssrval'   : SSE on testing set if provided
          -  'R2val'    : R2 on testing set if provided
    """
    data, debug = alamopy.data, alamopy.debug

    # patched together validation data check
    if "xval" in kwargs.keys():
        vargs = (kwargs["xval"], kwargs["zval"])
    else:
        vargs = ()

    xdata, zdata, xvaldata, zvaldata = setupData(
        data, debug, xdata, zdata, vargs, kwargs
    )
    manageArguments(xdata, zdata, data, debug, kwargs)

    data["results"] = {}

    writeCustomALAMOOptions(kwargs)  # New Custom Options MENGLE

    # Cross Validation
    if debug["loo"]:
        q2 = []
        if debug["outkeys"] and debug["expandoutput"]:
            q2 = {}

        # size = len(xdata) - 1
        data["opts"]["ndata"] = data["opts"]["ndata"] - 1
        kwargValidation = debug["validation"]
        kwargSaveTrace = debug["savetrace"]
        if kwargValidation:
            kwargNvaldata = data["opts"]["nvaldata"]

        data["opts"]["nvaldata"] = 1
        debug["validation"] = True
        debug["savetrace"] = False
        for i in range(0, len(xdata)):
            cvxdata = [x for y, x in enumerate(xdata) if y != i]
            cvzdata = [x for y, x in enumerate(zdata) if y != i]
            alamopy.almwriter(
                data, debug, (cvxdata, cvzdata, [xdata[i][:]], [zdata[i][:]]), kwargs
            )

            # Calling ALAMO
            if not debug["mock"]:
                os.system(
                    debug["almloc"]
                    + " "
                    + str(data["stropts"]["almname"])
                    + " > logscratch"
                )

            data["results"] = {}
            readTraceFile([xdata[i][:], zdata[i][:]], data, debug)

            if debug["outkeys"] and debug["expandoutput"]:
                for k in data["results"]["R2"].keys():
                    if k not in q2.keys():
                        q2[k] = [float(data["results"]["R2val"][k])]
                    else:
                        q2sub = q2[k]
                        q2sub.append(float(data["results"]["R2val"][k]))
                        q2[k] = q2sub
            else:
                q2.append(float(data["results"]["R2val"]))
            cleanFiles(data, debug)

        if debug["outkeys"] and debug["expandoutput"]:
            data["results"]["Q2"] = {}
            for k in q2.keys():
                Q2 = np.mean(q2[k])
                data["results"]["Q2"][k] = Q2
                print("%s: Running cross validation LOO, evaluated Q2:%f" % (k, Q2))
        else:
            Q2 = np.mean(q2)
            data["results"]["Q2"] = Q2
            print("Running cross validation LOO, evaluated Q2:%f" % Q2)

        del data["opts"]["nvaldata"]
        debug["validation"] = kwargValidation
        debug["savetrace"] = kwargSaveTrace
        if kwargValidation:
            data["opts"]["nvaldata"] = kwargNvaldata
        data["opts"]["ndata"] = data["opts"]["ndata"] + 1
    elif debug["lmo"] > 0:
        q2 = []
        if debug["outkeys"] and debug["expandoutput"]:
            q2 = {}

        kwargNdata = data["opts"]["ndata"]
        kwargValidation = debug["validation"]
        kwargSaveTrace = debug["savetrace"]
        if kwargValidation:
            kwargNvaldata = data["opts"]["nvaldata"]

        debug["validation"] = True
        debug["savetrace"] = False
        numOfFolds = debug["lmo"]
        print(xdata)
        if numOfFolds > len(xdata):
            raise Exception(
                "Number of Cross validation \
                            folds exceeds the number of data points"
            )

        # size = len(xdata)
        sizeOfFolds = int(len(xdata) / numOfFolds)
        r = len(xdata) % numOfFolds
        remS = 0
        remE = 1

        for i in range(numOfFolds):
            if i < r + 1:
                remS = i
                remE = i + 1
            cvvalxdata = xdata[remS + sizeOfFolds * i : sizeOfFolds * (i + 1) + remE]
            cvvalzdata = zdata[remS + sizeOfFolds * i : sizeOfFolds * (i + 1) + remE]
            if i == 0:
                cvxdata = xdata[sizeOfFolds * (i + 1) + remE :]
                cvzdata = zdata[sizeOfFolds * (i + 1) + remE :]
            else:
                cvxdata = np.concatenate(
                    [
                        xdata[0 : remS + sizeOfFolds * i],
                        xdata[sizeOfFolds * (i + 1) + remE :],
                    ]
                )
                cvzdata = np.concatenate(
                    [
                        zdata[0 : remS + sizeOfFolds * i],
                        zdata[sizeOfFolds * (i + 1) + remE :],
                    ]
                )

            data["opts"]["nvaldata"] = len(cvvalxdata)
            data["opts"]["ndata"] = len(cvxdata)

            alamopy.almwriter(
                data, debug, (cvxdata, cvzdata, cvvalxdata, cvvalzdata), kwargs
            )

            # Calling ALAMO
            if not debug["mock"]:
                os.system(
                    debug["almloc"]
                    + " "
                    + str(data["stropts"]["almname"])
                    + " > logscratch"
                )

            data["results"] = {}
            expandOutput(xdata, zdata, [cvvalxdata, cvvalzdata], data, debug)
            readTraceFile([cvvalxdata, cvvalzdata], data, debug)

            if debug["outkeys"] and debug["expandoutput"]:
                for k in data["results"]["R2"].keys():
                    if k not in q2.keys():
                        q2[k] = [float(data["results"]["R2val"][k])]
                else:
                    q2sub = q2[k]
                    q2sub.append(float(data["results"]["R2val"][k]))
                    q2[k] = q2sub
            else:
                q2.append(float(data["results"]["R2val"]))
            cleanFiles(data, debug)

        if debug["outkeys"] and debug["expandoutput"]:
            data["results"]["Q2"] = {}
            for k in q2.keys():
                Q2 = np.mean(q2[k])
                data["results"]["Q2"][k] = Q2
                print("%s: Running cross validation LMO, evaluated Q2:%f" % (k, Q2))
        else:
            Q2 = np.mean(q2)
            data["results"]["Q2"] = Q2
            print("Running cross validation LMO, evaluated Q2:%f" % Q2)

        del data["opts"]["nvaldata"]
        debug["validation"] = kwargValidation
        debug["savetrace"] = kwargSaveTrace
        if kwargValidation:
            data["opts"]["nvaldata"] = kwargNvaldata
        data["opts"]["ndata"] = kwargNdata

    # Write alamo file
    if debug["validation"]:
        alamopy.almwriter(data, debug, (xdata, zdata, xvaldata, zvaldata), kwargs)
    else:
        alamopy.almwriter(data, debug, (xdata, zdata), kwargs)

    # Call alamo from the terminal
    if not debug["mock"]:
        if debug["showalm"]:
            os.system(debug["almloc"] + " " + str(data["stropts"]["almname"]))
        else:
            writethis("Calling ALAMO now:\n")
            os.system(
                debug["almloc"]
                + " "
                + str(data["stropts"]["almname"])
                + " > logscratch"
            )

    # Check to see if additional data was sampled and add it
    if "sampler" in kwargs.keys():
        xdata, zdata = checkForSampledData(data, debug)

    # calculate additional statistics
    expandOutput(xdata, zdata, vargs, data, debug)

    # Open the trace file and pull appropriate results
    readTraceFile(vargs, data, debug)

    # write python file of regressed model
    alamopy.almpywriter(data)
    if debug["cvfun"]:
        alamopy.almcvwriter(data)

    # add <>alm.py to results dict
    data["results"]["pymodel"] = data["stropts"]["almname"].split(".")[0] + "alm"

    if debug["loo"] or debug["lmo"] > 0:
        Q2, R2, diff = 0, 0, 0
        if debug["outkeys"]:
            for k in data["results"]["R2"].keys():
                R2 = data["results"]["R2"][k]
                if k in data["results"]["Q2"].keys():
                    Q2 = data["results"]["Q2"][k]
                    diff = float(R2) - float(Q2)
                    if Q2 < 0.5:
                        print(
                            "%s: Q2 suggests this is not a predictive model, \
                              Q2: %f, R2: %s"
                            % (k, Q2, R2)
                        )
                    elif diff < 0.3:
                        print(
                            "%s: The difference of R2-Q2 is  %f. This is an acceptable \
                              difference for predictability, Q2: %f, R2: %s"
                            % (k, diff, Q2, R2)
                        )
                    else:
                        print(
                            "%s: The difference of R2-Q2 is %f. The surrogate model is \
                              not able to predict the data reliably, Q2: %f, R2: %s"
                            % (k, diff, Q2, R2)
                        )

        else:
            R2 = data["results"]["R2"]
            Q2 = data["results"]["Q2"]
            diff = float(R2) - float(Q2)

            if Q2 < 0.5:
                print(
                    "Q2 suggests this is not a predictive model, Q2: %f, R2: %s"
                    % (Q2, R2)
                )
            elif diff < 0.3:
                print(
                    "The difference of R2-Q2 is  %f. This is an acceptable difference for \
                      predictability, Q2: %f, R2: %s"
                    % (diff, Q2, R2)
                )
            else:
                print(
                    "The difference of R2-Q2 is %f. The surrogate model is not able to \
                      predict the data reliably, Q2: %f, R2: %s"
                    % (diff, Q2, R2)
                )

    cleanFiles(data, debug, pywrite=True, **kwargs)

    return data["results"]


# Data Management


def setupData(data, debug, xdata, zdata, vargs, kwargs):
    """ [xdata, zdata, xvaldata, zvaldata] = 
            setupData(data, debug, xdata,zdata, vargs, kwargs)

      Checks inputted data and resturctures the data for the .alm file

      Args:
          data/debug: shared default options for .alm file
          xdata (numpy.array or list[real])
          zdata (numpy.array or list[real)
          vargs:  Validation data
          kwargs: Additional options may be specified and will be applied
                  to the .alm

    """

    xvaldata, zvaldata = list(), list()
    checkinput(data, debug, xdata, zdata, vargs, kwargs)
    xdata, zdata = getTrainingData(xdata, zdata, data, debug)
    if len(vargs) > 0:
        xvaldata, zvaldata = getValidationData(vargs, data, debug)
    else:
        debug["validation"] = False
    getlabels(data, debug, kwargs)

    return xdata, zdata, xvaldata, zvaldata


def checkinput(data, debug, xdata, zdata, vargs, kwargs):
    """Check the input data into doalamo for errors.

      Args:
          data/debug: shared default options for .alm file
          xdata (numpy.array or list[real])
          zdata (numpy.array or list[real)
          vargs:  Validation data
          kwargs: Additional options may be specified and will be applied
                  to the .alm

    """

    # t = np.zeros([1, 1])
    kk = kwargs.keys()
    # First check vargs for validation set viability
    if len(vargs) > 0 and len(vargs) != 2:
        raise almerror.AlamoInputError(
            "Validation Xdata and Zdata must be"
            "specified in the following form: "
            "alamopy.doalamo(x,z,xval,zval)"
        )

    for arg in kk:
        if arg not in data["pargs"]["opts"] + data["pargs"]["stropts"] + data["pargs"][
            "lstopts"
        ] + data["pargs"]["set4"] + debug["pargs"] + ["xlabels", "zlabels"] + [
            "xval",
            "zval",
        ]:
            raise almerror.AlamoInputError(
                "The following argument given to"
                "doalamo() is not understood: {}".format(arg)
            )

    # read keys for debug
    for key in debug["pargs"]:
        if key in kk:
            debug[key] = kwargs[key]

    if "almname" in kk:
        data["stropts"]["almname"] += ".alm"


def getTrainingData(xdata, zdata, data, debug):
    """ Structure data for training the model. Modifies data['opts']
    
        Args:
        xdata (numpy.array or list[real])
        zdata (numpy.array or list[real)
        data:  shared alamo data options
        debug: Additional options may be specified and will be applied
                to the .alm
    """

    dshape = np.shape(xdata)
    if len(dshape) == 0:
        debug["traindata"] = False
    elif len(dshape) == 1:
        data["opts"]["ndata"] = np.shape(xdata)[0]
        data["opts"]["ninputs"] = 1
    else:
        data["opts"]["ndata"] = np.shape(xdata)[0]
        data["opts"]["ninputs"] = np.shape(xdata)[1]
    xdata = np.asarray(xdata)

    # Check training data
    if len(np.shape(zdata)) == 1:
        zdata = np.reshape(zdata, (data["opts"]["ndata"], 1))
        data["opts"]["noutputs"] = 1
    else:
        data["opts"]["noutputs"] = np.shape(zdata)[1]
    if np.shape(zdata)[0] != data["opts"]["ndata"]:
        almerror("p1")
    elif np.shape(xdata)[0] != data["opts"]["ndata"]:
        almerror("p1")
    zdata = np.asarray(zdata)
    return xdata, zdata


def getValidationData(vargs, data, debug):
    """ Structure data for validating the model. Modifies data['opts']

        Args:
        vargs: validation data valxdata, valzdata
        data:  shared alamo data options
        debug: Additional options may be specified and will be applied
                to the .alm
    """

    if vargs != ():
        debug["validation"] = True
        xvaldata = vargs[0]
        zvaldata = vargs[1]
        temp = np.shape(xvaldata)
        data["opts"]["nvaldata"] = temp[0]
        if len(np.shape(zvaldata)) == 1:
            zvaldata = np.reshape(zvaldata, (data["opts"]["nvaldata"], 1))
        if temp[1] != data["opts"]["ninputs"]:
            writethis("Number of input variables inconsistent between x and xval")
            almerror("p2")
        temp = np.shape(zvaldata)
        if temp[0] != data["opts"]["nvaldata"] or temp[1] != data["opts"]["noutputs"]:
            writethis("Problem with zval")
            almerror("p2")
        return xvaldata, zvaldata


# Function/variable labels


def getlabels(data, debug, kwargs):
    """ Creates labels for data and output. Modifies data['labs']. \
        Makes labels if no labels are given.

        Args:
          data:  shared alamo data options
          debug: Additional options may be specified and will be applied
                  to the .alm
          vargs: validation data valxdata, valzdata
    """
    # This function generates labels if they are not provided
    # Check to see if labels have been specified before we generate them
    if "xlabels" in kwargs.keys():
        data["labs"]["savexlabels"] = kwargs["xlabels"]
    else:
        # Make savexlabels
        temp = list()
        for i in range(data["opts"]["ninputs"]):
            temp.append("x" + str(i + 1))
        data["labs"]["savexlabels"] = temp

    if "zlabels" in kwargs.keys():
        data["labs"]["savezlabels"] = kwargs["zlabels"]
    else:
        # Make savezlabels
        temp = list()
        for i in range(data["opts"]["noutputs"]):
            temp.append("z" + str(i + 1))
        data["labs"]["savezlabels"] = temp

    # create temp labels for alm file
    # This avoids having to check user labels for alm rules
    data["labs"]["xlinks"] = [0 for i in range(data["opts"]["ninputs"])]
    data["labs"]["zlinks"] = [0 for i in range(data["opts"]["noutputs"])]
    makelabs(data, debug, "ninputs")
    makelabs(data, debug, "noutputs")


def makelabs(data, debug, param):
    """
    Constructs labels for alamo

      Args:
        data:  shared alamo data options
        debug: Additional options may be specified and will be applied
                to the .alm
        param = 'ninputs' or 'noutputs
    """

    temp = list([])
    for i in range(data["opts"][param]):
        if param == "ninputs":
            data["labs"]["xlinks"][i] = ["a", "b"]
            temp.append("almx" + str(i + 1) + "d")
            data["labs"]["xlinks"][i][0] = "almx" + str(i + 1) + "d"
            data["labs"]["xlinks"][i][1] = data["labs"]["savexlabels"][i]
            key = "xlabels"
        elif param == "noutputs":
            data["labs"]["zlinks"][i] = ["a", "b"]
            temp.append("almz" + str(i + 1) + "d")
            data["labs"]["zlinks"][i][0] = "almz" + str(i + 1) + "d"
            data["labs"]["zlinks"][i][1] = data["labs"]["savezlabels"][i]
            key = "zlabels"
    data["lstopts"][key] = temp


# Extra capabilities
group_list = []


def addBasisGroups(groups):
    """
    format: index_num  Type_of_function Member_indices(indice of the input) <pow>
    types: LIN, LOG, EXP, SIN, COS, MONO, MULTI2, MULTI3, RATIO, RBf, CUST, CONST
    Only one CONST
    One input: MONO, EXP, LOG, SIN, COS - integer from 1- NINPUTs, (-1) - all inputs
    multiple input: Multi2, Multi3, ratio
    Need <pow> for MONO, Multi2, Multi3, Ratio, (-1111) - all powers
    """
    print("Adding groups", groups)
    for g in groups:
        if len(g) == 1:
            addBasisGroup(g[0])
        elif len(g) == 2:
            addBasisGroup(g[0], g[1])
        elif len(g) == 3:
            addBasisGroup(g[0], g[1], g[2])


def addBasisGroup(type_of_function, input_indices="", powers=""):
    """
        Include the checks
    """
    global group_list
    group_index = len(group_list) + 1
    if type_of_function == "CONST":
        group_list.append("%i %s" % (group_index, type_of_function))
    elif type_of_function in ["LIN", "LOG", "EXP", "SIN", "COS"]:
        if input_indices == "":
            input_indices = "-1"
        group_list.append("%i %s %s" % (group_index, type_of_function, input_indices))
    elif type_of_function in ["MONO", "MULTI2", "MULTI3", "RATIO"]:
        if input_indices == "":
            input_indices = "-1"
        if powers == "":
            powers = "-1111"
        group_list.append(
            "%i %s %s %s" % (group_index, type_of_function, input_indices, powers)
        )
    elif type_of_function in ["GRP", "RBF", "CUST"]:
        group_list.append("%i %s %s" % (group_index, type_of_function, input_indices))


def clearBasisGroups():
    global group_list
    group_list = []


basis_constraint_list = []


def addBasisConstraints(groups_constraints):
    """
    NMT (no more than), ATL (at least), REQ (required if main group), XCL (exclude)
    format: group-id output_id constraint_type integer_parameter

    """
    global group_constraints, basis_constraints
    print("Adding basis selection constraint", group_constraints)

    for b in basis_constraints:
        addBasisConstraint(b[0], b[1], b[2], b[3])


def addBasisConstraint(group_id, output_id, constraint_type, intParam):
    global basis_constraint_list
    basis_constraint_list.append(
        "%i %i %s %i" % (group_id, output_id, constraint_type, intParam)
    )


def clearBasisConstraints():
    global basis_constraint_list
    basis_constraint_list = []


extrapxmin = None
extrapxmax = None
surface_constraint_list = []


def addCustomConstraints(constraint_list, **kwargs):
    """
        constraint args = CRTOL, CRNINITIAL, CRMAXITER, CRNVIOL, CRNTRIALS
    """
    print("Adding constraints", constraint_list)

    global extrapxmin, extrapxmax, surface_constraint_list
    if "extrapxmin" in kwargs.keys():
        extrapxmin = kwargs["extrapxmin"]
    if "extrapxmax" in kwargs.keys():
        extrapxmax = kwargs["extrapxmax"]
    for c in constraint_list:
        surface_constraint_list.append(c)


custom_fxn_list = []


def addCustomFunctions(fxn_list):
    print("Adding Functions", fxn_list)

    global custom_fxn_list

    for f in fxn_list:
        custom_fxn_list.append(f)
    print(custom_fxn_list)


def clearCustomFunctions():
    global custom_fxn_list
    custom_fxn_list = []


def writeCustomALAMOOptions(kwargs):
    global custom_fxn_list, surface_constraint_list, extrapxmax, extrapxmin, \
        group_list, basis_constraint_list
    name = "almopt.txt"

    if "almopt" in kwargs.keys():
        name = kwargs["almopt"]

    with open(name, "w") as r:
        if len(custom_fxn_list) > 0:
            r.write("\nNCUSTOMBAS %d\n" % len(custom_fxn_list))
            r.write("BEGIN_CUSTOMBAS\n")
            for c_fxn in custom_fxn_list:
                r.write("%s\n" % c_fxn)
            r.write("END_CUSTOMBAS\n\n")

        if len(surface_constraint_list) > 0:
            if extrapxmin is not None:
                r.write("extrapxmin %d\n" % extrapxmin)
            if extrapxmax is not None:
                r.write("extrapxmax %d\n" % extrapxmax)

        r.write("crncustom %i\n" % len(surface_constraint_list))
        i = 1
        r.write("BEGIN_CUSTOMCON\n")
        for c in surface_constraint_list:
            r.write("%s\n" % (c))
            i += 1
        r.write("END_CUSTOMCON\n")

        if len(group_list) > 0:
            r.write("\nNGROUPS %d\n" % len(group_list))
            r.write("BEGIN_GROUPS\n")
            for group in group_list:
                r.write("%s\n" % group)
            r.write("END_GROUPS\n\n")

        if len(basis_constraint_list) > 0:
            r.write("BEGIN_GROUPCON\n")
            for b in basis_constraint_list:
                r.write("%s\n" % b)
            r.write("END_GROUPCON\n\n")


# Argument management


def manageArguments(xdata, zdata, data, debug, kwargs):
    """ Parse additional input options
      The 'pargs' library is used to keep track of options a user has availible
      descriptions of the dictionaries data, and debug are given in shared.py
      Multiple keys used to make writing the .alm file easier
        
      Args:
        xdata (numpy.array or list[real])
        zdata (numpy.array or list[real)
        data:  shared alamo data options
        debug: Additional options may be specified and will be applied
                to the .alm
    """
    parseKwargs(data, debug, kwargs)

    # Check to see if a simwrapper should be built
    if debug["simwrap"] or "simulator" in kwargs.keys():
        buildSimWrapper(data, debug)

    # Specific check to see if the labels of the response variables
    # should be used in the output dictionary
    # This is important for systematic testing vs. single model input
    if debug["outkeys"]:
        # outkeys are specified to be used
        if data["opts"]["noutputs"] > 1:
            # 'Must use outkeys for multiple outputs'
            writethis("outkeys set to TRUE for multiple outputs")
            debug["outkeys"] = True

    # Construct xmin and xmax vector based on training data if not provided
    if "xmin" not in kwargs.keys():
        constructXBounds(xdata, zdata, data, debug)


def parseKwargs(data, debug, kwargs):
    """ Parse keyword arguments

      Args:
        data:  shared alamo data options
        debug: Additional options may be specified and will be applied
                to the .alm
        kwargs: keyword arguments
    """
    for arg in kwargs.keys():
        if arg in data["pargs"]["opts"]:
            data["opts"][arg] = kwargs[arg]
        elif arg in data["pargs"]["lstopts"]:
            data["lstopts"][arg] = list()
            for term in list([kwargs[arg]]):
                data["lstopts"][arg].append(term)
        elif arg in data["pargs"]["stropts"]:
            data["stropts"][arg] = kwargs[arg]
        elif arg in data["pargs"]["set4"]:
            data["set4"][arg] = kwargs[arg]
        elif arg in debug["pargs"]:
            debug[arg] = kwargs[arg]
        else:
            if arg not in (["xlabels", "zlabels", "xval", "zval"]):
                sys.stdout.write("Problem with option : " + arg)
                almerror("p3")


def buildSimWrapper(data, debug):
    """ Builds an executable simulator to sample for data 
      
      Args:
        data:  shared alamo data options
        debug: Additional options may be specified and will be applied
                to the .alm
    """

    if not isinstance(data["stropts"]["simulator"], type("string")):
        try:
            data["stropts"]["simulator"] = data["stropts"]["simulator"].__name__
        except Exception:
            raise almerror.AlamoInputError(
                "Simulator must be provided as a string"
                "and obey ALAMOs simulator conventions"
                "OR must be a python function whose name"
                "can be obtained via .__name__"
            )
        data["stropts"]["simulator"] = alamopy.wrapwriter(data["stropts"]["simulator"])
        debug["simwrap"] = True


def constructXBounds(xdata, zdata, data, debug):
    """ Construct xmin,xmax and zmin, zmax for alamo if none are given
        
      Args:
        xdata (numpy.array or list[real])
        zdata (numpy.array or list[real)
        data:  shared alamo data options
        debug: Additional options may be specified and will be applied
                to the .alm
    """
    writethis(
        "min and max values of inputs are not provided, \
              they will be calculated from the training data\n"
    )
    xmin = ""
    xmax = ""
    if data["opts"]["ninputs"] > 1:
        for i in range(data["opts"]["ninputs"]):
            tn = debug["bignum"]
            tx = -1 * debug["bignum"]
            for j in range(data["opts"]["ndata"]):
                if float(xdata[j][i]) < float(tn):
                    tn = xdata[j][i]
                if float(xdata[j][i]) > float(tx):
                    tx = xdata[j][i]
            xmin = xmin + str(tn) + " "
            xmax = xmax + str(tx) + " "
    else:
        tn = debug["bignum"]
        tx = -1 * debug["bignum"]
        for j in range(data["opts"]["ndata"]):
            if float(xdata[j]) < float(tn):
                tn = xdata[j]
            if float(xdata[j]) > float(tx):
                tx = xdata[j]
        xmin = xmin + str(tn) + " "
        xmax = xmax + str(tx) + " "
    data["set4"]["xmax"] = xmax
    data["set4"]["xmin"] = xmin


def checkForSampledData(data, debug):
    """ Check to see if data has been sampled and update ndata
        
      Args:
        data:  shared alamo data options
        debug: Additional options may be specified 
               and will be applied to the .alm
    """
    lst_name = data["stropts"]["almname"].split(".")[0] + ".lst"
    with open(lst_name) as infile, open("awkres", "w") as outfile:
        copy = False
        for line in infile:
            if "Errors on observed data points" in line.strip():
                copy = True
            elif "Maximum absolute errors" in line.strip():
                copy = False
            elif copy:
                outfile.write(line)
    f = open("awkres")
    lf = f.read()
    f.close()
    lf2 = lf.split("\n")
    lf2 = lf2[1:-1]
    sys.stdout.write(
        "Updating number of training points from "
        + str(data["opts"]["ndata"])
        + " to "
        + str(len(lf2))
        + "\n"
    )
    data["opts"]["ndata"] = len(lf2)
    xdata = np.zeros([data["opts"]["ndata"], data["opts"]["ninputs"]])
    zdata = np.zeros([data["opts"]["ndata"], data["opts"]["noutputs"]])
    for i in range(len(lf2)):
        lf3 = lf2[i].split(" ")
        while "" in lf3:
            lf3.remove("")
        for j in range(data["opts"]["ninputs"]):
            xdata[i][j] = float(lf3[j])
        for j in range(data["opts"]["noutputs"]):
            zdata[i][j] = float(lf3[data["opts"]["ninputs"] + j])
    deletefile("awkres")
    return xdata, zdata


def expandOutput(xdata, zdata, vargs, data, debug):
    """ Expand output to validation metrics and labels

        Args:
          data/debug: shared default options for .alm file
          xdata (numpy.array or list[real])
          zdata (numpy.array or list[real)
          vargs:  Validation data

    """
    if debug["expandoutput"]:
        data["results"]["xdata"] = xdata
        data["results"]["zdata"] = zdata
        data["results"]["xlabels"] = data["labs"]["savexlabels"]
        data["results"]["zlabels"] = data["labs"]["savezlabels"]
    # try:
    #     # import sympy
    #     from sympy.parsing.sympy_parser import parse_expr
    #     from sympy import symbols, lambdify
    # except Exception:
    #     writethis('Cannot import sympy')

    if debug["expandoutput"]:
        data["results"]["model"] = {}
        data["results"]["f(model)"] = {}
    else:
        for key in list(
            [
                "model",
                "f(model)",
                "ssr",
                "R2",
                "size",
                "rmse",
                "nbas",
                "totaltime",
                "olrtime",
                "miptime",
                "clrtime",
                "othertime",
                "version",
                "status",
                "madp",
                "numolr",
                "nummip",
                "numclr",
                "ninputs",
            ]
        ):
            data["results"][key] = collections.OrderedDict()
        if len(vargs) > 0:
            data["results"]["ssrval"] = collections.OrderedDict()
            data["results"]["R2val"] = collections.OrderedDict()
            data["results"]["rmseval"] = collections.OrderedDict()
            data["results"]["madpval"] = collections.OrderedDict()
    if debug["loo"]:
        data["results"]["Q2"] = data["results"]["Q2"]


# External File management


def readTraceFile(vargs, data, debug):
    """ Read the alamo trace file to read in the model and metrics

      Args:
          data/debug: shared default options for .alm file
          vargs:  Validation data

    """

    trace_file = data["stropts"]["tracefname"]
    # currentDirectory = os.getcwd()
    trace_str = trace_file  # currentDirectory + "/" + trace_file
    try:
        lf = open(trace_str).read()
    except (IOError, FileNotFoundError) as err:
        if debug["mock"]:
            data["results"]["clrtime"] = "0"
            data["results"]["size"] = "6"
            data["results"]["numolr"] = "16960"
            data["results"]["othertime"] = "0.8799995E-01"
            data["results"]["olrtime"] = "0.10800002"
            data["results"]["miptime"] = "0"
            data["results"]["version"] = "2018.4.3"
            data["results"]["status"] = "0"
            data["results"]["R2"] = "1"
            data["results"]["numclr"] = "0"
            data["results"]["nummip"] = "0"
            data["results"]["ssr"] = "0.169E-21"
            data["results"]["pymodel"] = "cam6alm"
            data["results"]["totaltime"] = "0.1760001"
            data["results"]["rmse"] = "0.255E-11"
            data["results"]["madp"] = "0.814E-09"
            data["results"][
                "model"
            ] = "  z1 = 3.9999999999884194856747 * x1^2 \
                 - 3.9999999999873385725380 * x2^2 - 2.0999999999876837186719 \
                 * x1^4 + 3.9999999999879496392907 * x2^4 + 0.33333333333014281141260 \
                 * x1^6 + 1.0000000000008837375276 * x1*x2"
            data["results"]["nbas"] = "15"

            if debug["expandoutput"]:
                data["results"]["ssrval"] = 0
                data["results"]["R2val"] = 0
                data["results"]["rmseval"] = 0
                data["results"]["madpval"] = 0
            return
        else:
            raise almerror.AlamoError(
                'Cannot read from trace file "{}": {}'.format(trace_str, err)
            )

    try:
        # import sympy
        from sympy.parsing.sympy_parser import parse_expr
        from sympy import symbols, lambdify
    except Exception:
        writethis("Cannot import sympy")

    lf2 = lf.split("\n")
    lf2_ind = 0  # ENGLE Allows for multiple writings to trace.trc file 5/30/19

    dict_out_str = (
        "#filename, NINPUTS, NOUTPUTS, INITIALPOINTS, OUTPUT, SET, "
        "INITIALIZER, SAMPLER, MODELER, BUILDER, GREEDYBUILD, "
        "BACKSTEPPER, GREEDYBACK, REGULARIZER, SOLVEMIP, SSEOLR, SSE, "
        "RMSE, R2, ModelSize, BIC, RIC, Cp, AICc, HQC, MSE, SSEp, MADp, "
        "OLRTime, numOLRs, OLRoneCalls, OLRoneFails, OLRgsiCalls, OLRgsiFails, "
        "OLRdgelCalls, OLRdgelFails, OLRclrCalls, OLRclrFails, OLRgmsCalls, "
        "OLRgmsFails, CLRTime, numCLRs, MIPTime, NumMIPs, LassoTime, "
        "Metric1Lasso, Metric2Lasso, LassoSuccess, LassoRed, nBasInitAct, "
        "nBas, SimTime, SimData, TotData, NdataConv, OtherTime, NumIters, "
        "IterConv, TimeConv, Step0Time, Step1Time, Step2Time, TotalTime, "
        "AlamoStatus, AlamoVersion, Model"
    )

    lf2_ind = len(lf2) - 1 - lf2[::-1].index(dict_out_str)
    tkeys = lf2[lf2_ind].split(",")
    kl1 = list(
        [
            "ssr",
            "rmse",
            "R2",
            "size",
            "nbas",
            "totaltime",
            "olrtime",
            "miptime",
            "clrtime",
            "othertime",
            "version",
            "status",
            "madp",
            "numolr",
            "nummip",
            "numclr",
            "ninputs",
        ]
    )
    kl2 = list(
        [
            " SSE",
            " RMSE",
            " R2",
            " ModelSize",
            " nBasInitAct",
            " TotalTime",
            " OLRTime",
            " MIPTime",
            " CLRTime",
            " OtherTime",
            " AlamoVersion",
            " AlamoStatus",
            " MADp",
            " numOLRs",
            " NumMIPs",
            " numCLRs",
            " NINPUTS",
        ]
    )
    # Construct results for training data (&val if provided)
    ln = 1
    wlparam = data["opts"]["noutputs"]

    # initialize multiple output expanded dictionary
    if debug["expandoutput"]:
        data["results"]["f(model)"] = {}
        data["results"]["model"] = {}
        for i in range(len(kl1)):
            data["results"][kl1[i]] = {}
        if len(vargs) > 0:
            for i in ["ssrval", "R2val", "rmseval", "madpval"]:
                data["results"][i] = {}

    if len(vargs) > 0:
        wlparam = 2 * wlparam
    else:
        wlparam = wlparam + 1
    while ln < wlparam:
        lf3 = lf2[lf2_ind + ln].split(",")
        # Reapply the saved labels for the output
        model = lf3[tkeys.index(" Model")]
        #    for label in data['labs']['savexlabels']:
        for i in range(data["opts"]["ninputs"]):
            label = data["labs"]["xlinks"][i][0]
            # Now is a convenient time to collect information that will be used in the
            # confidence interval analysis
            model = model.replace(str(label), str(data["labs"]["xlinks"][i][1]))
        for i in range(data["opts"]["noutputs"]):
            label = data["labs"]["zlinks"][i][0]
            model = model.replace(str(label), str(data["labs"]["zlinks"][i][1]))
        # determine which output label to write
        # if debug['outkeys'] == True use olab as a key if not dont

        if debug["outkeys"]:
            olab = model.split("=")[0]
            olab = olab.replace(" ", "")
            # print data['results'].keys
            data["results"]["model"][olab] = model
            # Record tokenized model for each output
            data["results"]["f(model)"][olab] = lambdify(
                [symbols(data["labs"]["savexlabels"])],
                parse_expr(model.split("=")[1].replace("^", "**")),
                "numpy",
            )
        else:
            data["results"]["model"] = model
            data["results"]["f(model)"] = lambdify(
                [symbols(data["labs"]["savexlabels"])],
                parse_expr(model.split("=")[1].replace("^", "**")),
                "numpy",
            )

        if debug["expandoutput"]:
            if debug["outkeys"]:
                for i in range(len(kl1)):
                    data["results"][kl1[i]][olab] = lf3[tkeys.index(kl2[i])]
                # Check for validation set
                if len(vargs) > 0:
                    lf3 = lf2[lf2_ind + 2].split(",")
                    data["results"]["ssrval"][olab] = lf3[tkeys.index(" SSE")]
                    data["results"]["R2val"][olab] = lf3[tkeys.index(" R2")]
                    data["results"]["rmseval"][olab] = lf3[tkeys.index(" RMSE")]
                    data["results"]["madpval"][olab] = lf3[tkeys.index(" MADp")]
            else:
                for i in range(len(kl1)):
                    data["results"][kl1[i]] = lf3[tkeys.index(kl2[i])]
                # Check for validation set
                if len(vargs) > 0:
                    lf3 = lf2[lf2_ind + 2].split(",")
                    data["results"]["ssrval"] = lf3[tkeys.index(" SSE")]
                    data["results"]["R2val"] = lf3[tkeys.index(" R2")]
                    data["results"]["rmseval"] = lf3[tkeys.index(" RMSE")]
                    data["results"]["madpval"] = lf3[tkeys.index(" MADp")]
        else:
            if debug["outkeys"]:
                data["results"]["ssr"][olab] = lf3[tkeys.index(kl2[0])]
            else:
                data["results"]["ssr"] = lf3[tkeys.index(kl2[0])]
        ln = ln + 1


def cleanFiles(data, debug, pywrite=False, **kwargs):
    """ Removes intermediate files

      Args:
          data/debug: shared default options for .alm file
          vargs:  Validation data

    """
    # Delete files
    if debug["mock"]:
        try:
            deletefile("temp.alm")
            if pywrite:
                for z in alamopy.data["zlabels"]:
                    deletefile("%s.py" % z)
            return
        except Exception:
            pass

    if not debug["savepyfcn"]:
        for z in alamopy.data["results"]["zlabels"]:
            deletefile("%s.py" % z)

    if not debug["savescratch"]:
        deletefile(
            str(data["stropts"]["almname"])
            + " "
            + str(data["stropts"]["almname"]).split(".")[0]
            + ".lst"
        )
    if not debug["savetrace"]:
        deletefile(data["stropts"]["tracefname"])
    if not debug["saveopt"]:
        global custom_fxn_list, surface_constraint_list, extrapxmin, extrapxmax
        custom_fxn_list = []
        surface_constraint_list = []
        extrapxmax = None
        extrapxmin = None
        if "almopt" in kwargs.keys():
            deletefile(kwargs["almopt"])

    if debug["simwrap"]:
        deletefile("simwrapper.py")


def get_alamo_version():
    x = [0, 1]
    z = [0, 1]

    res = alamo(x, z, saveopt=False)  # xval=xival, zval=zival, mock=True)

    try:
        deletefile("logscratch")
        deletefile("almopt.txt")
        deletefile("z1.py")
    except Exception:
        pass
    return res["version"]



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""HELMET Plotting capabilities"""

import scipy.stats as stats
import numpy as np
import math

import matplotlib.pyplot as plt
from matplotlib import cm
import matplotlib
from mpl_toolkits.mplot3d import Axes3D
import pandas as pd

from . import (
    DataImport,
    DataManipulation,
    BasisFunctions,
    parseGAMS,
    AncillaryEquations
)

font = {"size": 12}
matplotlib.rc("font", **font)
not_parsed = True

CVValuesn, SNDValues = [], []
critT, critD, critP, acc, R, M, Rm = [0, 0, 0, 0, 0, 0, 0]
molecule = ""
triple = 0

global props
props = ["PVT", "CV", "CP", "SND"]
markers = [".", ",","o","v","^", "<", ">", "1","2", "3","4","8","s", "p","P","*","x","X","D","d","|","_"];


def molData(fluidData, Dmolecule, RVal):
    """
    Shared data about the molecule and ideas gas constant R
    """
    global critT, critP, critD, M, triple, acc, R, Rm, molecule
    molecule = Dmolecule
    (critT, critP, critD, M, triple, acc) = fluidData
    R = RVal
    Rm = RVal


# View Ancillary Regressions
def viewAnc():
    """ Plots all the ancillary equations for saturated density and vapor pressure"""
    plotPV()
    plotDV()
    plotDL()


def plotPV():
    """Plot vapor pressure"""
    global critT, critP, critD, M, triple, acc, R, Rm, S

    DataImport.PV(molecule)
    Values = DataImport.Values
    DataToWrite = []
    T = []
    D = []
    for x in Values:
        D.append(float(x[0]))
        T.append(float(x[1]))
        manVal = DataManipulation.PV(x)
        DataToWrite.append(manVal)

    Dcalc = []

    for i, t in zip(DataToWrite, T):
        PVf = AncillaryEquations.getPV()
        PV = np.exp(PVf.f(i[0], i[1])) * critP
        Dcalc.append(PV)

    fig = plt.figure()
    fig.suptitle("Vapor Pressure Data", fontsize=20, fontweight="bold")
    ax = fig.add_subplot(111)
    ax.scatter(T, D, s=10, edgecolor="r", facecolor="w", marker="o", label="data")
    ax.scatter(T, Dcalc[:], s=10, c="g", marker="o", label="calculated")
    ax.set_ylabel("Pressure (MPa)", fontsize=20)
    ax.set_xlabel("Temperature (K)", fontsize=20)
    ax.legend(loc="upper right", fontsize=20)
    plt.show()


def plotDV():
    """Plot saturated vapor density"""
    global critT, critP, critD, M, triple, acc, R, Rm, S

    DataImport.DV(molecule)
    Values = DataImport.DVValues
    DataToWrite = []
    T = []
    D = []
    for x in Values:
        D.append(float(x[0]))
        T.append(float(x[1]))

        manVal = DataManipulation.DV(x)
        DataToWrite.append(manVal)

    Dcalc = []

    DV = AncillaryEquations.getDV()
    for i, t in zip(DataToWrite, T):
        Dc = float(critD)
        Ts = 1 - float(t) / float(critT)
        Dcalc.append(np.exp(DV.f(Ts)) * Dc)

    fig = plt.figure()
    fig.suptitle("Saturated Vapor Density", fontsize=20, fontweight="bold")
    ax = fig.add_subplot(111)
    ax.scatter(T, D, s=10, edgecolor="r", facecolor="w", marker="o", label="data")
    ax.scatter(T, Dcalc[:], s=10, c="g", marker="o", label="calculated")
    ax.set_ylabel("Density (dM)", fontsize=20)
    ax.set_xlabel("Temperature (K)", fontsize=20)
    ax.legend(loc="upper right", fontsize=20)
    plt.show()


def plotDL():
    """Plot saturate liquid density"""
    global critT, critP, critD, M, triple, acc, R, Rm, molecule
    DataImport.DL(molecule)
    Values = DataImport.DLValues

    DataToWrite = []
    T = []
    D = []
    for x in Values:
        D.append(float(x[0]))
        T.append(float(x[1]))
        manVal = DataManipulation.DL(x)
        DataToWrite.append(manVal)

    Dcalc = []

    for i, t in zip(DataToWrite, T):
        Dc = critD
        DL = AncillaryEquations.getDL()
        Dcalc.append((DL.f(float(i[0])) + 1) * (Dc))

    fig = plt.figure()
    fig.suptitle("Saturated Liquid Density", fontsize=20, fontweight="bold")
    ax = fig.add_subplot(111)
    ax.scatter(T, D, s=10, edgecolor="r", facecolor="w", marker="o", label="data")
    ax.scatter(T, Dcalc[:], s=10, c="g", marker="o", label="calculated")
    ax.set_ylabel("Density (dM)", fontsize=20)
    ax.set_xlabel("Temperature (K)", fontsize=20)
    ax.legend(loc="upper right", fontsize=20)
    plt.show()


# View Data
def viewData():
    """
    View imported data
    """
    global props

    if "PVT" in props:
        plotPVT()
    if "CV" in props:
        plotCV()
    if "CP" in props:
        plotCP()
    if "SND" in props:
        plotSND()


def plotPVT():
    """
    Plot Pressure-Volume-Temperature data
    """
    DataImport.PVT(molecule)
    PVTValues = DataImport.PVTValues

    PVTnp = np.asarray(PVTValues)
    T = PVTnp[:, 2]
    D = PVTnp[:, 1]
    P = PVTnp[:, 0]

    fig = plt.figure()
    fig.suptitle("Pressure Data", fontsize=20, fontweight="bold")

    ax = fig.add_subplot(111, projection="3d")
    ax.scatter(T, D, P, s=20, c=D, marker="o", label="data")

    # Plot Labels
    ax.set_zlabel("Pressure", fontsize=12)
    ax.set_ylabel("Density", fontsize=12)
    ax.set_xlabel("Temperature (K)", fontsize=12)
    ax.legend(loc="upper right", fontsize=15)
    plt.show()


def plotCV():
    """
    Plot isochoric heat capacity
    """
    DataImport.CV(molecule)
    CVValues = DataImport.CVValues
    DT = []
    for x in CVValues:
        Tau = float(critT) / float(x[1])
        Delta = float(x[0]) / float(critD)
        CVValuesn.append((Delta, Tau, float(x[2])))
        DT.append((Delta, Tau))
    CVnp = np.asarray(CVValuesn)

    fig = plt.figure()
    ax = fig.add_subplot(111, projection="3d")
    ax.scatter(
        CVnp[:, 1], CVnp[:, 0], CVnp[:, 2], s=20, c="b", marker="o", label="NIST data"
    )
    fig.suptitle("Isochoric Heat Capacity Data", fontsize=20, fontweight="bold")

    ax.legend(loc="upper right", fontsize=15)
    ax.set_zlabel("Isochoric Heat Capacity", fontsize=12)
    ax.set_ylabel("Reduced Density", fontsize=12)
    ax.set_xlabel("Inverse Reduced Temperature", fontsize=12)
    plt.show()


def plotCP():
    """
    Plot isobaric heat capacity
    """
    DataImport.CP(molecule)
    CPValues = DataImport.CPValues
    CPValuesn = []
    DT = []
    for x in CPValues:
        Theta = float(critT) / float(x[1])
        De = float(x[0]) / float(critD)
        z = float(x[2])
        CPValuesn.append((De, Theta, z, float(x[0])))
        DT.append((De, Theta))

    CPnp = np.asarray(CPValuesn)
    T = CPnp[:, 1]
    D = CPnp[:, 0]
    CP = CPnp[:, 2]

    fig = plt.figure()
    ax = fig.add_subplot(111, projection="3d")
    ax.scatter(T, D, CP, s=20, c=T, marker="o", label="NIST data")
    fig.suptitle("Isobaric Heat Capacity Data", fontsize=20, fontweight="bold")
    ax.legend(loc="upper right")
    ax.set_zlabel("$C_p$", fontsize=12)
    ax.set_ylabel(r"$\delta$", fontsize=12)
    ax.set_xlabel("$ \\tau $", fontsize=12)

    plt.show()


def plotSND():
    """
    Plot speed of sound data
    """
    DataImport.SND(molecule)
    Values = DataImport.SNDValues
    WS = []

    for x in Values:
        Theta = float(critT) / float(x[1])
        Delta = float(x[0]) / float(critD)
        z = float(x[2])
        WS.append([Theta, Delta, z])

    SNDnp = np.asarray(WS)
    T = SNDnp[:, 0]
    D = SNDnp[:, 1]
    w = SNDnp[:, 2]

    w = (w ** 2) / R / 1000 * M / T * float(critT)

    fig = plt.figure()
    ax = fig.add_subplot(111, projection="3d")

    ax.scatter(T, D, w, s=25, c="r", marker="o", label="NIST data")
    ax.legend(loc="upper right", fontsize=15)
    fig.suptitle("Speed of Sound Data", fontsize=20, fontweight="bold")
    ax.set_zlabel("Dimensionless Speed of Sound", fontsize=12)
    ax.set_ylabel(r"$\delta$", fontsize=15)
    ax.set_xlabel("$\\tau$", fontsize=15)

    plt.show()


# View Regressed Equation
def sseCombo(lstFile=None, plot = False, report=False, surface=cm.coolwarm):
    """Plot regressed equation and data. Calculates statistical anlaysis metrics"""
    Y = []
    Beta = []

    parseGAMS.parser(lstFile)
    Y = parseGAMS.indexes[0]
    Beta = [float(b) for b in parseGAMS.betas[0]]

    BasisFunctions.formCustomBasis()
    parseGAMS.writeEquation(Y, Beta)

    saveFig = False
    showGraph = plot

    objZ, objCV, objCP, objSND = 0, 0, 0, 0

    if 'PVT' in props:
        objZ = ssePVT(Y, Beta, saveFig, showGraph, report)
    if 'CV' in props:
        objCV = sseCV(Y, Beta, saveFig, showGraph, report)
    if 'CP' in props:
        objCP = sseCP(Y, Beta, saveFig, showGraph, report)
    if 'SND' in props:
        objSND = sseSND(Y, Beta, saveFig, showGraph, report)

    if showGraph:
        HelmetSurface(Y, Beta, showGraph, surface)


def ssePVT(PVT1=[], PVT1Vals=[], saveFig=False, show=True, report=False):
    """Plots and metrics for Pressure-volume-temperature data"""
    Y, Beta = PVT1, PVT1Vals
    BasisFunctions.formCustomBasis()

    DT = []
    PlaceHolder = []
    Z = []

    DataImport.PVT(molecule)
    PVTValues = DataImport.PVTValues

    DataToWrite = []
    for x in PVTValues:
        manVal = DataManipulation.PVT(x)
        DataToWrite.append(manVal)

    for x in DataToWrite:
        Pval = BasisFunctions.drdRes(x[0], x[1], Y, Beta)
        PlaceHolder.append(Pval)
        Z.append(x[2])

    PVTnp = np.asarray(PVTValues)
    P = []
    DT = []
    for x in PVTValues:
        P.append(float(x[0]))
        T = float(critT) / float(x[2])
        D = float(x[1]) / float(critD)
        DT.append((D, T))

    anp = np.asarray(DT)
    D = anp[:, 0]
    T = anp[:, 1]
    Zvar = np.var(Z)
    Ze = [(y - x) for x, y in zip(Z, PlaceHolder)]
    Z = [(y - x) ** 2 for x, y in zip(Z, PlaceHolder)]
    PlaceHolder = [1 + x for x in PlaceHolder]
    PlaceHolder = [
        y * R / 1000 * float(x[2]) * float(x[1]) for x, y in zip(PVTValues, PlaceHolder)
    ]

    Ta = PVTnp[:, 2]
    Pa = np.array(P)


    delPcalc = (Pa - PlaceHolder) / Pa
    AAD = np.mean(delPcalc)

    critVal = BasisFunctions.drdRes(1, 1, Y, Beta)
    d2rdv = BasisFunctions.d2rdRes(1, 1, Y, Beta)
    d3rdv = BasisFunctions.d3rdRes(1, 1, Y, Beta)
    if report:
        print("Critical Values", critVal)
    CritPressureCalc = (critVal + 1) * R / 1000 * float(critT) * float(critD)
    if report:
        print("Critical Pressures", float(critP), CritPressureCalc)

    CritFirstDeriv = 1 + 2 * critVal + d2rdv
    CritSecondDeriv = 2 * critVal + 4 * d2rdv + d3rdv

    if report:
        print("First Deriv", 0, CritFirstDeriv)
        print("Second Deriv", 0, CritSecondDeriv)
    errors = [(m - x) ** 2 for m, x in zip(Pa, PlaceHolder)]

    sseCalc = sum(errors)
    avgY = sum(Pa) / len(Pa)
    sseTotal = sum((Pa - avgY) ** 2)
    R2calc = 1 - sseCalc / sseTotal
    meanRes = np.mean(Pa - PlaceHolder)
    residuals = [m - x for m, x in zip(Pa, PlaceHolder)]

    S2 = sum((residuals - meanRes) ** 2) / (len(residuals) - 1)
    S20 = sum(np.array(residuals) ** 2) / (len(residuals) - 1)
    tscore = meanRes / np.sqrt(S2 + S20)

    if report:
        print("Pressure ----------------------")
        print("SSE for Fit %f" % sseCalc)
        print("Average residual %f" % np.mean(Pa - PlaceHolder))
        print("R2 calculated is %f" % R2calc)
        RMSE = np.sqrt(np.mean(errors))
        print("RMSE calculated is %f" % RMSE)
        print("AAD is %f, %f, %f" % (min(delPcalc), AAD, max(delPcalc)))
        print("t value %f, df %f" % (tscore, len(residuals)))
        print(stats.ttest_1samp(residuals, 0))

    if show or saveFig:
        fig2 = plt.figure()
        fig2.suptitle('%s Pressure Volume Temperature Regression'% molecule)

        # Pressure Parity Plot
        ax2 = fig2.add_subplot(131)
        # ax2.set_title("%s PVT Regression" % molecule)
        ax2.set_xlabel("Observed pressure (MPa)")
        ax2.set_ylabel("Calculated pressure (MPa)")

        ax2.plot(
            [0, max(max(Pa), max(PlaceHolder))],
            [0, max(max(Pa), max(PlaceHolder))],
            c="b",
            label="$R^2 = %.3f$" % R2calc,
        )
        ax2.set_xlim([0, 120])
        ax2.set_ylim([0, 120])
        ax2.scatter(
            Pa,
            PlaceHolder,
            c="w",
            s=10,
            edgecolor="k",
            label="Developed equation\n (multivariable)"
        )
        ax2.legend(loc="upper left")


        # Pressure Data Plot
        ax = fig2.add_subplot(132)
        ax.scatter(
            Ta, Pa, s=25, c="r", marker="o", edgecolor="k", label="%s data" % molecule
        )
        ax.scatter(
            Ta,
            PlaceHolder,
            s=20,
            c="b",
            marker="o",
            edgecolor="k",
            label="Developed equation"
        )
        # ax.set_title('Data Fit');
        # ax1.set_title('Residuals')
        ax.set_ylabel("Pressure (MPa)")
        ax.set_xlabel("Temperature (K)")
        ax.legend(loc="upper right")

        #  Residuals Plot
        ax4 = fig2.add_subplot(133)
        ax4.scatter(T, Pa - PlaceHolder, c=residuals, edgecolor="k")
        ax4.set_xlabel("Temperature (K)")
        ax4.set_ylabel("Residuals")

        # 3d Plot
        fig3 = plt.figure()
        fig3.suptitle('%s Pressure Volume Temperature Regression'% molecule)
        ax3 = fig3.add_subplot(111, projection="3d")
        ax3.scatter(D, T, Pa, edgecolor="k", label="Data")
        ax3.scatter(D, T, PlaceHolder, edgecolor="k", label="Regression")
        ax3.set_xlabel("Reduced Density")
        ax3.set_ylabel("Inverse Reduced Temperature")
        ax3.set_zlabel("Pressure (MPa)")

        # figAAD = plt.figure()
        # axAAD = figAAD.add_subplot(111)
        # PVTAAD = [ (x-y)/x for x, y in zip(Pa,PlaceHolder)]
        # SAAD = [ x[0] for x in zip(Sources,PVTAAD) if abs(x[1]) >1]
        # TAAD = [ x[0] for x in zip(T,PVTAAD) if abs(x[1]) >1]
        # PVTAAD = [ x[1] for x in zip(T,PVTAAD) if abs(x[1]) >1]

        # df = pd.DataFrame(dict(x=TAAD, y=PVTAAD, label=SAAD))
        # groups = df.groupby('label')

        # i = 0
        # for name, group in groups:
        #     axAAD.scatter(group.x, group.y, label=name, marker=markers[i])
        #     i+=1
        # # axAAD.scatter(TAAD, PVTAAD, label=SAAD)
        # axAAD.legend(loc="right")


        if saveFig:
            fig2.savefig("%sPics/%sPVT.eps" % (molecule, molecule))
            fig3.savefig("%sPics/%sPVTR2.eps" % (molecule, molecule))
        if show:
            plt.show()

    if report:
        print("VARZ: ", Zvar)
    # l-1 norm
    sseZ1 = [abs(x) for x in Ze]
    # l-2 norm
    sseZ2 = [x ** 2 for x in Ze]
    # print "SSEZ2: ", np.sum(sseZ1)
    return [np.sum(sseZ1) / Zvar, np.sum(sseZ2) / Zvar, np.sum(sseZ1), np.sum(sseZ2)]


def sseCV(Y =[], Beta=[], saveFig=False, show=True, report=False):
    """Plots and metrics for isochoric heat capacity"""
    DT = []
    CVValuesn = []
    PlaceHolder = []

    global DataToWrite
    CVValues = []
    try:
        DataImport.CV(molecule)
        CVValues = DataImport.CVValues
    except Exception:
        print("Error plotting CV")
        return
    else:
        pass

    CVValues = DataImport.CVValues
    DataToWrite = []
    for x in CVValues:
        manVal = DataManipulation.CV(x)
        DataToWrite.append(manVal)
    BasisFunctions.formCustomBasis()

    for x in DataToWrite:
        val = BasisFunctions.rTTRes(x[0], x[1], Y, Beta)
        itt_val = BasisFunctions.iTT(x[0], x[1])
        CVR = val + itt_val
        PlaceHolder.append(CVR)

    for x in CVValues:
        Tau = float(critT) / float(x[1])
        Delta = float(x[0])/float(critD)
        CVValuesn.append((Delta, Tau, float(x[2])))
        DT.append((Delta, Tau))

    PlaceHolder = [-z * R for T, z in zip(DT, PlaceHolder)]
    CVnp = np.asarray(CVValuesn)
    T = CVnp[:, 1]
    CV = CVnp[:, 2]

    if show or saveFig:
        fig = plt.figure()
        fig.suptitle('%s Isochoric Heat Capacity Regression'%molecule)
        ax = fig.add_subplot(121)
        ax.scatter(
            T, CV, s=10, c="b", edgecolors="k", marker="s", label="%s data" % molecule
        )
        ax.scatter(
            T,
            PlaceHolder,
            s=10,
            c="r",
            marker="o",
            edgecolors="k",
            label="Developed equation",
        )
        ax.legend(loc="upper left")
        ax.set_ylabel("Isochoric Heat Capacity")
        ax.set_xlabel("Temperature (K)")

        delCVcalc = (CV - PlaceHolder) / (CV)
        AAD = np.mean(delCVcalc)
        sseCalc = sum((CV - PlaceHolder) ** 2)
        avgY = sum(CV) / len(CV)
        sseTotal = sum((CV - avgY) ** 2)
        R2calc = 1 - sseCalc / sseTotal
        errors = (CV - PlaceHolder) ** 2

        meanRes = np.mean(CV - PlaceHolder)
        residuals = [m - x for m, x in zip(CV, PlaceHolder)]
        S2 = sum((residuals - meanRes) ** 2) / (len(residuals) - 1)
        S20 = sum(np.array(residuals) ** 2) / (len(residuals) - 1)
        t = meanRes / np.sqrt(S2 + S20)

        if report:
            print("Isochoric Heat Capacity --------------------------")
            print("SSE for Fit %f" % sseCalc)
            print("Average residual %f" % np.mean(CV - PlaceHolder))
            RMSE = np.sqrt(np.mean(errors))
            print("RMSE calculated is %f" % RMSE)
            # print "SSE total %f" % sseTotal;
            print("R2 calculated is %f" % R2calc)
            print("AAD is %f, %f, %f" % (min(delCVcalc), AAD, max(delCVcalc)))
            print("t value %f, df %i" % (t, len(residuals)))
            print(stats.ttest_1samp(residuals, 0))

        # Parity Plot
        # fig2 = plt.figure()
        ax2 = fig.add_subplot(122)
        ax2.set_xlabel(r"$C_V/(J/(mol\cdot K))$ observed")
        ax2.set_ylabel(r"$C_V/(J/(mol\cdot K))$ calculated")

        maxCV = math.ceil(max(max(CV),max(PlaceHolder)))
        ax2.plot([0, maxCV], [0, maxCV], label="$R^2 =  %.3f$" % R2calc, color="b")
        # ax2.plot([0,550], [0,550], label='$R^2 =  %.3f$' %R2calc, color='b')
        ax2.scatter(
            CV,
            PlaceHolder,
            edgecolors="k",
            facecolors="none",
            label="Developed equation",
        )
        # ax2.set_xlim([0,550])
        # ax2.set_ylim([0,550])
        ax2.legend(loc="upper left")

        # fig3 = plt.figure()
        # ax3 = fig3.add_subplot(111, projection='3d');
        # ax3.scatter(D, T, CV, edgecolor = 'k', label ='Data')
        # ax3.scatter(D,T, PlaceHolder, edgecolor='k', label='Regression')
        # ax3.set_xlabel('Delta');
        # ax3.set_ylabel('Tau');
        # ax3.set_zlabel('Pressure')

        if saveFig:
            fig.savefig("%sPics/%sCV.eps" % (molecule, molecule))
            # fig2.savefig('%sPics/%sCVR2.eps' %(molecule,molecule))
        if show:
            plt.show()

    CVR = [x / R for x in CV]
    PR = [x / R for x in PlaceHolder]
    # l-1 norm
    sseCV1 = [abs(x - y) for x, y in zip(CVR, PR)]
    # l-2 norm
    sseCV2 = [(x - y) ** 2 for x, y in zip(CVR, PR)]
    if report:
        print("VARCV: ", np.var(CVR))
    return [sum(sseCV1) / np.var(CVR), sum(sseCV2) / np.var(CVR)]


def sseCP(CP1=[], CP1Vals=[], saveFig=False, show=True, report=False):
    """Plots and calculates metrics for isobaric heat capacity"""
    global molecule
    Y = CP1
    Beta = CP1Vals

    BasisFunctions.formCustomBasis()
    DataImport.CP(molecule)

    CPValues = DataImport.CPValues
    CPValuesn = []
    DT = []

    rD = []
    rDD = []
    rTD = []
    PlaceHolder = []

    for x in CPValues:
        Theta = float(critT) / float(x[1])
        De = float(x[0]) / float(critD)
        z = float(x[2])
        CPValuesn.append((De, Theta, z, float(x[0])))
        DT.append((De, Theta))

    for x in DT:
        val = BasisFunctions.rTTRes(x[0], x[1], Y, Beta)
        itt_val = BasisFunctions.iTT(x[0], x[1])
        CPR = val + itt_val
        PlaceHolder.append(CPR)
        val = BasisFunctions.drdRes(x[0], x[1], Y, Beta)
        rD.append(val)
        rDD.append(BasisFunctions.d2rdRes(x[0], x[1], Y, Beta))
        rTD.append(BasisFunctions.dtrdtRes(x[0], x[1], Y, Beta))

    CPnp = np.asarray(CPValuesn)
    T = CPnp[:, 1]
    D = CPnp[:, 0]
    CP = CPnp[:, 2]

    dPD = [1 + 2 * y + z for x, y, z in zip(DT, rD, rDD)]  # 10 to 10^-2
    dPT = [1 + y - z for x, y, z in zip(DT, rD, rTD)]  # 10^-1

    CPcalc = [-x + (y ** 2) / z for x, y, z in zip(PlaceHolder, dPT, dPD)]
    CPcalc = [x * R for x in CPcalc]


    errors = (CP - CPcalc) ** 2
    delCPcalc = (CP - CPcalc) / CP
    AAD = np.mean(delCPcalc)
    sseCalc = sum((CP - CPcalc) ** 2)
    avgY = sum(CP) / len(CP)
    sseTotal = sum((CP - avgY) ** 2)
    R2calc = 1 - sseCalc / sseTotal

    meanRes = np.mean(CP - CPcalc)
    residuals = [m - x for m, x in zip(CP, CPcalc)]
    S2 = sum((residuals - meanRes) ** 2) / (len(residuals) - 1)
    S20 = sum(np.array(residuals) ** 2) / (len(residuals) - 1)
    t = meanRes / np.sqrt(S2 + S20)

    if report:
        print("Isobaric Heat Capacity ---------------------------")
        print("SSE for Fit %f" % sseCalc)
        print("Average residual %f" % np.mean(CP - CPcalc))
        RMSE = np.sqrt(np.mean(errors))
        print("RMSE calculated is %f" % RMSE)
        # print "SSE total %f" % sseTotal;
        print("R2 calculated is %f" % R2calc)
        print("AAD is %f, %f, %f" % (min(delCPcalc), AAD, max(delCPcalc)))
        print("t value %f, df %i" % (t, len(residuals)))
        print(stats.ttest_1samp(residuals, 0))


    if show or saveFig:

        fig2 = plt.figure()
        fig2.suptitle("%s Isobaric Heat Capacity Regression"%molecule)

        # Parity Plot
        ax2 = fig2.add_subplot(131)
        ax2.set_xlabel(r"$C_P/(J/(mol\cdot K))$ observed")
        ax2.set_ylabel(r"$C_P/(J/(mol\cdot K))$ calculated")
        minCP = math.floor(min(min(CP),min(CPcalc)))
        maxCP = math.ceil(max(max(CP),max(CPcalc)))
        ax2.plot([minCP, maxCP], [minCP, maxCP], label="$R^2 = %.3f$" % R2calc, color="b")
        ax2.set_xlim([minCP, maxCP])
        ax2.set_ylim([minCP, maxCP])
        ax2.scatter(
            CP,
            CPcalc,
            edgecolors="k",
            facecolors="none",
            label="Developed equation\n (multivariable)",
        )
        ax2.legend(loc="upper right")


        # Scatter Plot
        ax = fig2.add_subplot(132)
        ax.scatter(
            T,
            CP,
            s=15,
            c="r",
            marker="s",
            edgecolors="k",
            label="%s data (NIST)" % molecule,
        )
        ax.scatter(
            T,
            CPcalc,
            s=15,
            c="b",
            marker="o",
            edgecolors="k",
            label="Developed equation",
        )
        ax.set_ylabel(r"$C_P/(J/(mol\cdot K))$")
        ax.set_xlabel("Inverse Reduced Temperature")
        ax.legend(loc="upper right")
        ax.set_xlim((0.8, 3.4))
        ax.set_ylim((0, 900))


        ax4 = fig2.add_subplot(133)

        ax4.scatter(
            D,
            CP,
            s=15,
            c="r",
            marker="s",
            edgecolors="k",
            label="%s data (NIST)" % molecule,
        )
        ax4.scatter(
            D,
            CPcalc,
            s=15,
            c="b",
            marker="o",
            edgecolors="k",
            label="Developed equation",
        )
        ax4.set_ylabel(r"$C_P/(J/(mol\cdot K))$")
        ax4.set_xlabel("Reduced Density")
        ax4.legend(loc="upper right")
        # ax4.set_title("%s Isobaric Heat Capacity Regression" % molecule)
        ax.legend(loc="upper right")


        fig3 = plt.figure()
        ax3 = fig3.add_subplot(111, projection="3d")
        ax3.scatter(D, T, CP, edgecolor="k", label="Data")
        ax3.scatter(D, T, CPcalc, edgecolor="k", label="Regression")
        ax3.set_xlabel("Reduced Density")
        ax3.set_ylabel("Inverse Reduced Temperature")
        ax3.set_zlabel("Pressure")
        ax3.set_title("%s Isobaric Heat Capacity Data" % molecule)
        ax3.legend()

        if saveFig:
            fig2.savefig("%sPics/%sCP.eps" % (molecule, molecule))
        if show:
            plt.show()

    CPR = [x / R for x in CP]
    # l -1 norm
    sseCP1 = [abs(x / R - y / R) for x, y in zip(CP, CPcalc)]
    # l-2  norm
    sseCP2 = [(x / R - y / R) ** 2 for x, y in zip(CP, CPcalc)]

    if report:
        print("VARCP: ", np.var(CPR))
    return [sum(sseCP1) / np.var(CPR), sum(sseCP2) / np.var(CPR)]


def sseSND(SND1=[], SND1Vals=[], saveFig=False, show=True, report=False):
    """Plots and calculates metrics for speed of sound """
    Y = SND1
    Beta = SND1Vals

    BasisFunctions.formCustomBasis()
    DataImport.SND(molecule)
    Values = DataImport.SNDValues

    DT = []
    WS = []

    rD = []
    rDD = []
    rTD = []
    rTT = []
    PlaceHolder = []
    SNDValues = []

    for x in Values:
        Theta = float(critT) / float(x[1])
        D = float(x[0]) / float(critD)
        z = float(x[2])
        SNDValues.append((D, float(x[1]), z, float(x[0])))
        DT.append((D, Theta))
        WS.append(z)

    SNDnp = np.asarray(SNDValues)
    T = SNDnp[:, 1]
    D = SNDnp[:, 0]
    w = SNDnp[:, 2]

    for x in DT:
        val = BasisFunctions.rTTRes(x[0], x[1], Y, Beta)
        itt_val = BasisFunctions.iTT(x[0], x[1])
        # was not negative, but .iTT returns - itt
        CPR = val + itt_val
        PlaceHolder.append(CPR)
        val = BasisFunctions.drdRes(x[0], x[1], Y, Beta)
        rD.append(val)
        rDD.append(BasisFunctions.d2rdRes(x[0], x[1], Y, Beta))
        rTD.append(BasisFunctions.dtrdtRes(x[0], x[1], Y, Beta))
        rTT.append(BasisFunctions.rTTRes(x[0], x[1], Y, Beta))

    dPD = [1 + 2 * x + y for x, y in zip(rD, rDD)]  # 10 to 10^-2
    dPT = [1 + x - y for x, y in zip(rD, rTD)]  # 10^-1
    Dwcalc = [x - (y ** 2) / z for x, y, z in zip(dPD, dPT, PlaceHolder)]
    Dwcalc = [x for x in Dwcalc]

    w = (w ** 2) / R / 1000 * M / T * float(critT)
    WS = [(x ** 2) / R / 1000 * M * y[1] / float(critT) for x, y in zip(WS, DT)]
    We = WS
    Wes = Dwcalc

    WSn = np.asarray(WS)
    Dw = np.asarray(Dwcalc)

    delWcalc = [(d - w) / d for d, w in zip(WS, Dwcalc)]
    AAD = np.mean(delWcalc)
    sseCalc = sum((WSn - Dw) ** 2)
    avgY = sum(WSn) / len(WSn)
    sseTotal = sum((WSn - avgY) ** 2)
    R2calc = 1 - sseCalc / sseTotal
    errors = [(d - w) ** 2 for d, w in zip(WS, Dwcalc)]
    residual = [d - w for d, w in zip(WS, Dwcalc)]

    meanRes = np.mean(residual)
    residuals = [m - x for m, x in zip(WS, Dwcalc)]
    S2 = sum((residuals - meanRes) ** 2) / (len(residuals) - 1)
    S20 = sum(np.array(residuals) ** 2) / (len(residuals) - 1)
    t = meanRes / np.sqrt(S2 + S20)

    if report:
        print("Speed of Sound ----------------------------")
        print("SSE for Fit %f" % sseCalc)
        print("Average residual %f" % np.mean(residual))
        RMSE = np.sqrt(np.mean(errors))
        print("RMSE calculated is %f" % RMSE)
        # print "SSE total %f" % sseTotal;
        print("R2 calculated is %f" % R2calc)
        print("AAD is %f, %f, %f" % (min(delWcalc), AAD, max(delWcalc)))
        print("t value %f, df %i" % (t, len(residuals)))
        print(stats.ttest_1samp(residuals, 0))

    if show or saveFig:
        fig2 = plt.figure()
        fig2.suptitle("%s Speed of Sound Regression"%molecule)

        # Parity Plot
        ax2 = fig2.add_subplot(121)
        ax2.set_xlabel("Speed of sound observed")
        ax2.set_ylabel("Speed of sound calculated")
        minSND = math.floor(min(min(WS),min(Dwcalc)))
        maxSND = math.ceil(max(max(WS),max(Dwcalc)))
        ax2.plot([minSND, maxSND], [minSND, maxSND], label="$R^2 = %.3f$" % R2calc, color="b")
        ax2.set_xlim([minSND, maxSND])
        ax2.set_ylim([minSND, maxSND])
        ax2.scatter(
            WS,
            Dwcalc,
            facecolors="none",
            edgecolors="k",
            label="Developed equation\n (multivariable)",
        )
        ax2.legend(loc="upper right")

        ax = fig2.add_subplot(122)

        ax.scatter(
            T, WS, s=25, c="r", marker="s", edgecolors="k", label="%s data" % molecule
        )
        ax.scatter(
            T,
            Dwcalc,
            s=20,
            c="b",
            marker="o",
            edgecolors="k",
            label="Developed equation",
        )
        ax.set_xlabel("Inverse reduced temperature")
        ax.set_ylabel("Speed of sound")
        ax.legend(loc="upper right")
        ax.set_xlim(min(T)*0.8, max(T)*1.1)
        ax.set_ylim(min(WS)*0.8, max(WS)*1.1)

        fig3 = plt.figure()
        fig3.suptitle("%s Speed of Sound Regression"%molecule)
        ax3 = fig3.add_subplot(111, projection="3d")
        ax3.scatter(D, T, WS, edgecolor="k", label="Data")
        ax3.scatter(D, T, Dwcalc, edgecolor="k", label="Regression")
        ax3.set_xlabel("Reduced Density")
        ax3.set_ylabel("Inverse Reduced Temperature")
        ax3.set_zlabel("Speed of sound")

        if saveFig:
            fig2.savefig("%sPics/%sSND.eps" % (molecule, molecule))
            fig3.savefig("%sPics/%sSNDR2.eps" % (molecule, molecule))
        if show:
            plt.show()

    if report:
        print("VAR", np.var(We))
    # l-1 norm
    sseW1 = [abs(x - y) for x, y in zip(Wes, We)]
    # l-2 norm
    sseW2 = [(x - y) ** 2 for x, y in zip(Wes, We)]
    return [sum(sseW1) / np.var(We), sum(sseW2) / np.var(We)]


def HelmetSurface(Y=[], Beta=[], show = True, surface=cm.coolwarm):
    """Plots Helmholtz Surface"""
    global critT, critP, critD, M, triple, acc, R, Rm, molecule
    BasisFunctions.formCustomBasis()
    # print(Y, Beta)

    fig = plt.figure()
    fig.suptitle("%s Helmholtz Energy Surface"%molecule)
    ax = fig.gca(projection="3d")

    D = np.arange(0.1, 2, 0.005)
    T = np.arange(0.1, 2, 0.005)
    D, T = np.meshgrid(D, T)

    PlaceHolder = np.zeros((len(D), len(T)))

    for i in range(len(D)):
        for j in range(len(D)):
            # print i,j, D[i,j], T[i,j]
            val = BasisFunctions.arBY(D[i, j], T[i, j], Y, Beta)
            itt_val = BasisFunctions.idealBY(D[i, j], T[i, j], Y, Beta)
            Helm = val + itt_val
            PlaceHolder[i, j] = Helm

    ax.plot_surface(
        D, T, PlaceHolder, cmap=surface, linewidth=0, antialiased=False
    )

    ax.xaxis.labelpad = 5
    ax.yaxis.labelpad = 5
    ax.zaxis.labelpad = 5

    ax.set_xlabel("Reduced density")
    ax.set_ylabel("Inverse reduced temperature")
    ax.set_zlabel("Helmholtz energy")

    ax.set_xticks([0.25, 0.75, 1.25, 1.75])
    ax.set_yticks([0.25, 0.75, 1.25, 1.75])
    # ax.set_zticks([-2, 0, 2])

    plt.show()



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Example ideal parameter block for the VLE calucations for a
Benzene-Toluene-o-Xylene system.
"""

# Import Python libraries
import logging

# Import Pyomo libraries
from pyomo.environ import Param, NonNegativeReals, Set

# Import IDAES cores
from idaes.core import declare_process_block_class
from idaes.core.util.misc import extract_data

from ideal_prop_pack_VLE import IdealParameterData

# Some more inforation about this module
__author__ = "Jaffer Ghouse"
__version__ = "0.0.1"


# Set up logger
_log = logging.getLogger(__name__)


@declare_process_block_class("BTXParameterBlock")
class BTXParameterData(IdealParameterData):

    def build(self):
        '''
        Callable method for Block construction.
        '''
        super(BTXParameterData, self).build()

        self.component_list_master = Set(initialize=['benzene',
                                                     'toluene',
                                                     'o-xylene'])

        # Component list - a list of component identifiers
        # NOTE: User needs to update this list; can be a subset or
        # equal to the master component list
        self.component_list = Set(initialize=['benzene', 'toluene'])

        # List of components in each phase (optional)
        self.phase_comp = {"Liq": self.component_list,
                           "Vap": self.component_list}

        # List of phase equilibrium index
        self.phase_equilibrium_idx_master = Set(initialize=[1, 2, 3])

        self.phase_equilibrium_idx = Set(initialize=[1, 2])

        self.phase_equilibrium_list_master = \
            {1: ["benzene", ("Vap", "Liq")],
             2: ["toluene", ("Vap", "Liq")],
             3: ["o-xylene", ("Vap", "Liq")]}

        self.phase_equilibrium_list = \
            {1: ["benzene", ("Vap", "Liq")],
             2: ["toluene", ("Vap", "Liq")]}

        # Thermodynamic reference state
        self.pressure_ref = Param(mutable=True,
                                  default=101325,
                                  doc='Reference pressure [Pa]')
        self.temperature_ref = Param(mutable=True,
                                     default=298.15,
                                     doc='Reference temperature [K]')

        # Source: The Properties of Gases and Liquids (1987)
        # 4th edition, Chemical Engineering Series - Robert C. Reid
        pressure_crit_data = {'benzene': 48.9e5,
                              'toluene': 41e5,
                              'o-xylene': 37.3e5
                              }

        self.pressure_crit = Param(
            self.component_list,
            within=NonNegativeReals,
            mutable=False,
            initialize=extract_data(pressure_crit_data),
            doc='Critical pressure [Pa]')

        # Source: The Properties of Gases and Liquids (1987)
        # 4th edition, Chemical Engineering Series - Robert C. Reid
        temperature_crit_data = {'benzene': 562.2,
                                 'toluene': 591.8,
                                 'o-xylene': 630.3
                                 }

        self.temperature_crit = Param(
            self.component_list,
            within=NonNegativeReals,
            mutable=False,
            initialize=extract_data(temperature_crit_data),
            doc='Critical temperature [K]')

        # Gas Constant
        self.gas_const = Param(within=NonNegativeReals,
                               mutable=False,
                               default=8.314,
                               doc='Gas Constant [J/mol.K]')

        # Source: The Properties of Gases and Liquids (1987)
        # 4th edition, Chemical Engineering Series - Robert C. Reid
        mw_comp_data = {'benzene': 78.1136E-3,
                        'toluene': 92.1405E-3,
                        'o-xylene': 106.167e-3}

        self.mw_comp = Param(self.component_list,
                             mutable=False,
                             initialize=extract_data(mw_comp_data),
                             doc="molecular weight Kg/mol")

        # Constants for liquid densities
        # Source: Perry's Chemical Engineers Handbook
        #         - Robert H. Perry (Cp_liq)
        dens_liq_data = {('benzene', '1'): 1.0162,
                         ('benzene', '2'): 0.2655,
                         ('benzene', '3'): 562.16,
                         ('benzene', '4'): 0.28212,
                         ('toluene', '1'): 0.8488,
                         ('toluene', '2'): 0.26655,
                         ('toluene', '3'): 591.8,
                         ('toluene', '4'): 0.2878,
                         ('o-xylene', '1'): 0.69883,
                         ('o-xylene', '2'): 0.26113,
                         ('o-xylene', '3'): 630.33,
                         ('o-xylene', '4'): 0.27429}

        self.dens_liq_params = Param(
                self.component_list,
                ['1', '2', '3', '4'],
                mutable=False,
                initialize=extract_data(dens_liq_data),
                doc="Parameters to compute liquid densities")

        # Boiling point at standard pressure
        # Source: Perry's Chemical Engineers Handbook
        #         - Robert H. Perry (Cp_liq)
        bp_data = {('benzene'): 353.25,
                   ('toluene'): 383.95,
                   ('o-xylene'): 417.15}

        self.temperature_boil = Param(
                self.component_list,
                mutable=False,
                initialize=extract_data(bp_data),
                doc="Pure component boiling points at standard pressure [K]")

        # Constants for specific heat capacity, enthalpy
        # Sources: The Properties of Gases and Liquids (1987)
        #         4th edition, Chemical Engineering Series - Robert C. Reid
        #         Perry's Chemical Engineers Handbook
        #         - Robert H. Perry (Cp_liq)
        cp_ig_data = {('Liq', 'benzene', '1'): 1.29E5,
                      ('Liq', 'benzene', '2'): -1.7E2,
                      ('Liq', 'benzene', '3'): 6.48E-1,
                      ('Liq', 'benzene', '4'): 0,
                      ('Liq', 'benzene', '5'): 0,
                      ('Vap', 'benzene', '1'): -3.392E1,
                      ('Vap', 'benzene', '2'): 4.739E-1,
                      ('Vap', 'benzene', '3'): -3.017E-4,
                      ('Vap', 'benzene', '4'): 7.130E-8,
                      ('Vap', 'benzene', '5'): 0,
                      ('Liq', 'toluene', '1'): 1.40E5,
                      ('Liq', 'toluene', '2'): -1.52E2,
                      ('Liq', 'toluene', '3'): 6.95E-1,
                      ('Liq', 'toluene', '4'): 0,
                      ('Liq', 'toluene', '5'): 0,
                      ('Vap', 'toluene', '1'): -2.435E1,
                      ('Vap', 'toluene', '2'): 5.125E-1,
                      ('Vap', 'toluene', '3'): -2.765E-4,
                      ('Vap', 'toluene', '4'): 4.911E-8,
                      ('Vap', 'toluene', '5'): 0,
                      ('Liq', 'o-xylene', '1'): 3.65e4,
                      ('Liq', 'o-xylene', '2'): 1.0175e3,
                      ('Liq', 'o-xylene', '3'): -2.63,
                      ('Liq', 'o-xylene', '4'): 3.02e-3,
                      ('Liq', 'o-xylene', '5'): 0,
                      ('Vap', 'o-xylene', '1'): -1.585e-1,
                      ('Vap', 'o-xylene', '2'): 5.962e-1,
                      ('Vap', 'o-xylene', '3'): -3.443e-4,
                      ('Vap', 'o-xylene', '4'): 7.528E-8,
                      ('Vap', 'o-xylene', '5'): 0}

        self.cp_ig = Param(self.phase_list, self.component_list,
                           ['1', '2', '3', '4', '5'],
                           mutable=False,
                           initialize=extract_data(cp_ig_data),
                           doc="parameters to compute Cp_comp")

        # Source: The Properties of Gases and Liquids (1987)
        # 4th edition, Chemical Engineering Series - Robert C. Reid
        pressure_sat_coeff_data = {('benzene', 'A'): 4.202,
                                   ('benzene', 'B'): 1322,
                                   ('benzene', 'C'): -38.56,
                                   ('toluene', 'A'): 4.216,
                                   ('toluene', 'B'): 1435,
                                   ('toluene', 'C'): -43.33,
                                   ('o-xylene', 'A'): 4.233,
                                   ('o-xylene', 'B'): 1548,
                                   ('o-xylene', 'C'): -51.65}

        self.pressure_sat_coeff = Param(
            self.component_list,
            ['A', 'B', 'C'],
            mutable=False,
            initialize=extract_data(pressure_sat_coeff_data),
            doc="parameters to compute Cp_comp")

        # Source: The Properties of Gases and Liquids (1987)
        # 4th edition, Chemical Engineering Series - Robert C. Reid
        dh_vap = {'benzene': 3.377e4, 'toluene': 3.8262e4,
                  'o-xylene': 4.34584e4}

        self.dh_vap = Param(self.component_list,
                            mutable=False,
                            initialize=extract_data(dh_vap),
                            doc="heat of vaporization")



__all__ = [
    "Helmet",
    "initialize",
    "DataImport",
    "AncillaryEquations",
    #"Certainty",
    "GAMSWrite",
    "Plotting",
    "GAMSDataWrite",
    "parseGAMS",
    "BasisFunctions",
    "DataManipulation",
    "plotDL",
    "plotDV",
    "plotPV",
    "DL",
    "DV",
    "PV",
    "viewAnc",
    #"molData",
    "GenerateGDXGamsFiledtlmv",
]


from .Helmet import initialize
from .AncillaryEquations import DL, DV, PV
from .BasisFunctions import formCustomBasis
from .GAMSWrite import GenerateGDXGamsFiledtlmv
from .Plotting import viewAnc, plotDL, plotDV, plotPV


#!/usr/bin/python
##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
import numpy as np
import sys


def sixcamel(*x):
    x1, x2 = x
    t1 = np.multiply(4.0 - 2.1 * np.power(x1, 2) + np.divide(np.power(x1, 4), 3.0),
                     np.power(x1, 2))
    t2 = np.multiply(4 * np.power(x2, 2) - 4,
                     np.power(x2, 2))
    z = t1 + np.multiply(x1, x2) + t2
    return z


def ackley(*x):
    import numpy as np
    x1, x2 = x
    a = 20
    b = 0.2
    c = 2 * 3.14159
    z = -a * np.exp(-b * np.sqrt(0.5 * (x1**2 + x2**2))) \
        - np.exp(0.5 * (np.cos(c * x1) + np.cos(c * x2))) + a + np.exp(1)
    return z


def branin(*x):
    import numpy as np
    x1, x2 = x
    pi = 3.14159
    z = (x2 - (5.1 / (4 * pi**2)) * x1**2 + (5 / pi) * x1 - 6)**2  \
        + 10 * (1 - (1 / (8 * pi)) * np.cos(x1) + 10) + np.random.normal(0, 0.1)
    return z


if __name__ == '__main__':
    sys.stdout.write(' ALAMOpy example functions ')
    sys.stdout.write(' call functions with : ')
    sys.stdout.write(' examples.<name>')
    sys.stdout.write(' <name> = branin ')
    sys.stdout.write('          sixcamel ')
    sys.stdout.write('          ackley ')



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Tests for versioning
"""
# third-party
import pytest
# pkg
import idaes
from idaes import ver


def test_idaes_version():
    assert idaes.__version__


def test_ver_class():
    v = ver.Version(1, 2, 3)
    assert str(v) == '1.2.3'
    v = ver.Version(1, 2, 3, 'beta', 1)
    assert str(v) == '1.2.3.b1'
    v = ver.Version(1, 2, 3, 'development')
    assert str(v) == '1.2.3.dev'
    pytest.raises(ValueError, ver.Version, 1, 2, 3, 'howdy')


class MyVersionedClass(ver.HasVersion):
    def __init__(self):
        super(MyVersionedClass, self).__init__(1, 2, 3)


def test_has_version():
    x = MyVersionedClass()
    assert str(x.version) == '1.2.3'


def test_bump_version():
    v = ver.Version(1, 2, 3)
    assert tuple(v) == (1, 2, 3)
    v.micro += 1
    assert tuple(v) == (1, 2, 4)



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Property package for ideal VLE calucations for the methanol synthesis problem.
Correlations from Turkay and Grossmann paper. See Latex files for details.
"""

# Import Python libraries
import logging

# Import Pyomo libraries
from pyomo.environ import Constraint, log, NonNegativeReals, value, Var
from pyomo.opt import SolverFactory, TerminationCondition

# Import IDAES cores
from idaes.core import (declare_process_block_class,
                        StateBlockData,
                        StateBlock,
                        MaterialFlowBasis,
                        MaterialBalanceType,
                        EnergyBalanceType)
from idaes.core.util.initialization import (fix_state_vars,
                                            revert_state_vars,
                                            solve_indexed_blocks)
from idaes.core.util.exceptions import ConfigurationError

# Some more inforation about this module
__author__ = "Jaffer Ghouse"
__version__ = "0.0.1"


# Set up logger
_log = logging.getLogger(__name__)


class _IdealStateBlock(StateBlock):
    """
    This Class contains methods which should be applied to Property Blocks as a
    whole, rather than individual elements of indexed Property Blocks.
    """

    def initialize(blk, state_args=None,
                   hold_state=False, outlvl=0,
                   solver='ipopt', optarg={'tol': 1e-8}):
        """
        Initialization routine for property package.

        Keyword Arguments:
            state_args : Dictionary with initial guesses for the state vars
                         chosen. Note that if this method is triggered
                         through the control volume, and if initial guesses
                         were not provied at the unit model level, the
                         control volume passes the inlet values as initial
                         guess.The keys for the state_args dictionary are:

                         flow_mol : value at which to initialize component
                                    flows
                         pressure : value at which to initialize pressure
                         temperature : value at which to initialize temperature
                         mole_frac_comp: value at which to initialize the component
                                    mixture mole fraction
            outlvl : sets output level of initialization routine

                     * 0 = no output (default)
                     * 1 = return solver state for each step in routine
                     * 2 = include solver output infomation (tee=True)

            optarg : solver options dictionary object (default=None)
            solver : str indicating whcih solver to use during
                     initialization (default = 'ipopt')
            hold_state : flag indicating whether the initialization routine
                         should unfix any state variables fixed during
                         initialization (default=False).
                         - True - states varaibles are not unfixed, and
                                 a dict of returned containing flags for
                                 which states were fixed during
                                 initialization.
                        - False - state variables are unfixed after
                                 initialization by calling the
                                 relase_state method

        Returns:
            If hold_states is True, returns a dict containing flags for
            which states were fixed during initialization.
        """
        # Fix state variables if not already fixed
        flags = fix_state_vars(blk, state_args)

        # Set solver options
        if outlvl > 1:
            stee = True
        else:
            stee = False

        if optarg is None:
            sopt = {'tol': 1e-8}
        else:
            sopt = optarg

        opt = SolverFactory('ipopt')
        opt.options = sopt

        # ---------------------------------------------------------------------
        for k in blk.keys():

            blk[k].eq_total.deactivate()
            blk[k].eq_comp.deactivate()
            if (blk[k].config.defined_state is False):
                blk[k].eq_mol_frac_out.deactivate()
            if (blk[k].config.has_phase_equilibrium) or \
                    (blk[k].config.parameters.config.valid_phase ==
                        ('Liq', 'Vap')) or \
                    (blk[k].config.parameters.config.valid_phase ==
                        ('Vap', 'Liq')):
                blk[k].eq_Keq.deactivate()
                blk[k].eq_sum_mol_frac.deactivate()

        if (blk[k].config.has_phase_equilibrium) or \
                (blk[k].config.parameters.config.valid_phase ==
                    ('Liq', 'Vap')) or \
                (blk[k].config.parameters.config.valid_phase ==
                    ('Vap', 'Liq')):
            results = solve_indexed_blocks(opt, [blk], tee=stee)

            if outlvl > 0:
                if results.solver.termination_condition \
                        == TerminationCondition.optimal:
                    _log.info("Initialization step 1 for "
                              "{} completed".format(blk.name))
                else:
                    _log.warning("Initialization step 1 for "
                                 "{} failed".format(blk.name))

        else:
            if outlvl > 0:
                _log.info("Initialization step 1 for "
                          "{} skipped".format(blk.name))

        for k in blk.keys():
            blk[k].eq_total.activate()
            blk[k].eq_comp.activate()
            if (blk[k].config.has_phase_equilibrium) or \
                    (blk[k].config.parameters.config.valid_phase ==
                        ('Liq', 'Vap')) or \
                    (blk[k].config.parameters.config.valid_phase ==
                        ('Vap', 'Liq')):
                blk[k].eq_Keq.activate()
                blk[k].eq_sum_mol_frac.activate()

        results = solve_indexed_blocks(opt, [blk], tee=stee)

        if outlvl > 0:
            if results.solver.termination_condition \
                    == TerminationCondition.optimal:
                _log.info("Initialization step 2 for "
                          "{} completed".format(blk.name))
            else:
                _log.warning("Initialization step 2 for "
                             "{} failed".format(blk.name))

        for k in blk.keys():
            if (blk[k].config.defined_state is False):
                blk[k].eq_mol_frac_out.activate()
        # ---------------------------------------------------------------------
        # If input block, return flags, else release state
        if hold_state is True:
            return flags
        else:
            blk.release_state(flags)

        if outlvl > 0:
            if outlvl > 0:
                _log.info('{} Initialization Complete.'.format(blk.name))

    def release_state(blk, flags, outlvl=0):
        '''
        Method to relase state variables fixed during initialization.

        Keyword Arguments:
            flags : dict containing information of which state variables
                    were fixed during initialization, and should now be
                    unfixed. This dict is returned by initialize if
                    hold_state=True.
            outlvl : sets output level of of logging
        '''
        # Unfix state variables
        revert_state_vars(blk, flags)

        if outlvl > 0:
            if outlvl > 0:
                _log.info('{} State Released.'.format(blk.name))


@declare_process_block_class("IdealStateBlock",
                             block_class=_IdealStateBlock)
class StateBlockData(StateBlockData):
    """An example property package for ideal VLE."""

    def build(self):
        """Callable method for Block construction."""
        super(StateBlockData, self).build()

        # Check for valid phase indicator and consistent flags
        if self.config.has_phase_equilibrium and \
                self.config.parameters.config.valid_phase in ['Vap', 'Liq']:
            raise ConfigurationError("Inconsistent inputs. Valid phase"
                                     " flag not set to VL for the state"
                                     " block but has_phase_equilibrium"
                                     " is set to True.")
        self._make_state_vars()
        self._make_vars()
        if not self.config.has_phase_equilibrium and \
                self.config.parameters.config.valid_phase == "Liq":
            self._make_liq_phase_eq()

        if (self.config.has_phase_equilibrium) or \
                (self.config.parameters.config.valid_phase ==
                    ('Liq', 'Vap')) or \
                (self.config.parameters.config.valid_phase ==
                    ('Vap', 'Liq')):
            self._make_flash_eq()

        if not self.config.has_phase_equilibrium and \
                self.config.parameters.config.valid_phase == "Vap":
            self._make_vap_phase_eq()

    def get_material_flow_basis(self):
        return MaterialFlowBasis.molar

    def _make_state_vars(self):
        """List the necessary state variable objects."""
        self.flow_mol = Var(initialize=1.0,
                            domain=NonNegativeReals,
                            doc='Component molar flowrate [kgmol/s]')
        self.mole_frac_comp = Var(self._params.component_list,
                             bounds=(0, 1),
                             initialize=1.0 / len(self._params.component_list))
        self.pressure = Var(initialize=0.101325,
                            domain=NonNegativeReals,
                            doc='State pressure [MPa]')
        self.temperature = Var(initialize=3,
                               domain=NonNegativeReals,
                               doc='State temperature [100K]')

    def _make_vars(self):
        self.flow_mol_phase = Var(self._params.phase_list,
                                  initialize=0.5,
                                  bounds=(0, None))

        self.mole_frac_phase_comp = Var(self._params.phase_list,
                                   self._params.component_list,
                                   initialize=1 / len(self._params.component_list),
                                   bounds=(0, 1))

    def _make_liq_phase_eq(self):
        def rule_total_mass_balance(self):
            return self.flow_mol_phase['Liq'] == self.flow_mol
        self.eq_total = Constraint(rule=rule_total_mass_balance)

        def rule_comp_mass_balance(self, i):
            return self.flow_mol * self.mole_frac_comp[i] == \
                self.flow_mol_phase['Liq'] * self.mole_frac_phase_comp['Liq', i]
        self.eq_comp = Constraint(self._params.component_list,
                                  rule=rule_comp_mass_balance)

        if self.config.defined_state is False:
            # applied at outlet only
            self.eq_mol_frac_out = Constraint(expr=sum(self.mole_frac_comp[i]
                                              for i in self._params.component_list)
                                              == 1)

    def _make_vap_phase_eq(self):
        def rule_total_mass_balance(self):
            return self.flow_mol_phase['Vap'] == self.flow_mol
        self.eq_total = Constraint(rule=rule_total_mass_balance)

        def rule_comp_mass_balance(self, i):
            return self.flow_mol * self.mole_frac_comp[i] == \
                self.flow_mol_phase['Vap'] * self.mole_frac_phase_comp['Vap', i]
        self.eq_comp = Constraint(self._params.component_list,
                                  rule=rule_comp_mass_balance)

        if self.config.defined_state is False:
            # applied at outlet only
            self.eq_mol_frac_out = Constraint(expr=sum(self.mole_frac_comp[i]
                                              for i in self._params.component_list)
                                              == 1)

    def _make_flash_eq(self):
        self.vapor_pressure = Var(self._params.component_list,
                                  initialize=0.101325,
                                  doc="vapor pressure ",
                                  bounds=(0.01, None))

        def rule_total_mass_balance(self):
            return self.flow_mol_phase['Liq'] + \
                self.flow_mol_phase['Vap'] == self.flow_mol
        self.eq_total = Constraint(rule=rule_total_mass_balance)

        def rule_comp_mass_balance(self, i):
            return self.flow_mol * self.mole_frac_comp[i] == \
                self.flow_mol_phase['Liq'] * self.mole_frac_phase_comp['Liq', i] + \
                self.flow_mol_phase['Vap'] * self.mole_frac_phase_comp['Vap', i]
        self.eq_comp = Constraint(self._params.component_list,
                                  rule=rule_comp_mass_balance)

        def rule_mole_frac(self):
            return sum(self.mole_frac_phase_comp['Liq', i]
                       for i in self._params.component_list) -\
                sum(self.mole_frac_phase_comp['Vap', i]
                    for i in self._params.component_list) == 0
        self.eq_sum_mol_frac = Constraint(rule=rule_mole_frac)

        if self.config.defined_state is False:
            # applied at outlet only
            self.eq_mol_frac_out = Constraint(expr=sum(self.mole_frac_comp[i]
                                              for i in self._params.component_list)
                                              == 1)

        def rule_Keq(self, i):
            return self.mole_frac_phase_comp['Vap', i] * self.pressure == \
                self.vapor_pressure[i] * self.mole_frac_phase_comp['Liq', i]
        self.eq_Keq = Constraint(self._params.component_list, rule=rule_Keq)

        def rule_P_vap(self, j):
            return log(7500.6168 * self.vapor_pressure[j]) == \
                self._params.vapor_pressure_coeff[j, 'A'] - \
                (self._params.vapor_pressure_coeff[j, 'B'] /
                 (100*self.temperature - self._params.vapor_pressure_coeff[j, 'C']))
        self.eq_P_vap = Constraint(self._params.component_list, rule=rule_P_vap)

    def _density_mol(self):
        self.density_mol = Var(self._params.phase_list, doc="Molar density [kgmol/m^3]")

        def density_mol_calculation(self, p):
            if p == "Vap":
                return 10.0 * self.pressure == (self.density_mol[p] *
                                         self._params.gas_constant *
                                         self.temperature)
            elif p == "Liq":
                return self.density_mol[p] == 11.1  # kgmol/m3 # dummy value
        try:
            # Try to build constraint
            self.density_mol_calculation = Constraint(
                self._params.phase_list, rule=density_mol_calculation)
        except AttributeError:
            # If constraint fails, clean up so that DAE can try again later
            self.del_component(self.density_mol)
            self.del_component(self.density_mol_calculation)
            raise

    def get_material_flow_terms(self, p, j):
        """Create material flow terms for control volume."""
        if (p == "Vap") and (j in self._params.component_list):
            return self.flow_mol_phase['Vap'] * self.mole_frac_phase_comp['Vap', j]
        elif (p == "Liq") and (j in self._params.component_list):
            return self.flow_mol_phase['Liq'] * self.mole_frac_phase_comp['Liq', j]
        else:
            return 0

    def get_enthalpy_flow_terms(self, p):
        """Create enthalpy flow terms [MJ/s]."""
        if p == "Vap":
            return self.flow_mol_phase['Vap'] * self._params.Cp * self.temperature * 100
        elif p == "Liq":
            return self.flow_mol_phase['Liq'] * self._params.Cp * self.temperature * 100

    def get_material_density_terms(self, p, j):
        """Create material density terms."""
        if p == "Liq":
            if j in self._params.component_list:
                return self.density_mol[p] * self.mole_frac_phase_comp['Liq', j]
            else:
                return 0
        elif p == "Vap":
            if j in self._params.component_list:
                return self.density_mol[p] * self.mole_frac_phase_comp['Vap', j]
            else:
                return 0

    def get_enthalpy_density_terms(self, p):
        """Create enthalpy density terms."""
        if p == "Liq":
            return self.density_mol[p] * self._params.Cp * self.temperature
        elif p == "Vap":
            return (self.density_mol[p] * (
                    self._params.Cp - self._params.gas_const) *
                    self.temperature)

    def default_material_balance_type(self):
        return MaterialBalanceType.componentTotal

    def default_energy_balance_type(self):
        return EnergyBalanceType.enthalpyTotal

    def define_state_vars(self):
        """Define state vars."""
        return {"flow_mol": self.flow_mol,
                "mole_frac_comp": self.mole_frac_comp,
                "temperature": self.temperature,
                "pressure": self.pressure}

    def model_check(blk):
        """Model checks for property block."""
        # Check temperature bounds
        if value(blk.temperature) < blk.temperature.lb:
            _log.error('{} Temperature set below lower bound.'
                       .format(blk.name))
        if value(blk.temperature) > blk.temperature.ub:
            _log.error('{} Temperature set above upper bound.'
                       .format(blk.name))

        # Check pressure bounds
        if value(blk.pressure) < blk.pressure.lb:
            _log.error('{} Pressure set below lower bound.'.format(blk.name))
        if value(blk.pressure) > blk.pressure.ub:
            _log.error('{} Pressure set above upper bound.'.format(blk.name))



from .turbine_inlet import TurbineInletStage
from .turbine_outlet import TurbineOutletStage
from .turbine_stage import TurbineStage
from .valve_steam import SteamValve, ValveFunctionType
from .turbine_multistage import TurbineMultistage
from .feedwater_heater_0D import FWH0D, FWHCondensing0D 



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
# stdlib
import os
# package
from idaes.surrogate.helmet import Helmet


def main():
    # Push information for the molecule
    num_terms = 14

    Fluids = {"H2O": (647.096, 22.064, 17.8737279956, 18.015268, 273.16, 0.344)}

    molecule = "H2O"
    (critT, critP, critD, M, triple, acc) = Fluids[molecule]
    R = 8.314472  # J mol^-1 K^-1

    # Constants for a molecule
    Helmet.initialize(
        molecule=molecule,
        fluid_data=Fluids[molecule],
        filename=os.getcwd() + "/%s" % molecule,
        gamsname=os.getcwd() + "/%s" % molecule,
        props=["PVT", "CV", "CP", "SND"],
        sample=3,
    )

    # Prepare Ancillary Equations of sat liq/vapor density and vapor pressure
    Helmet.prepareAncillaryEquations()  # plot=True

    # View data used for regression
    Helmet.viewPropertyData()

    # Write GAMS Gdx data file and regression file
    Helmet.setupRegression(numTerms=num_terms, gams=True)

    # runs the GAMS gdx and regression file
    # Helmet.runRegression(gams=True)

    # View Results by importing the data
    # Helmet.viewResults("H2Omain.lst", plot=True)


if __name__ == "__main__":
    main()
